from rest_framework import viewsets

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, MultiIdentifierUpdateMixin
from api_fhir_r4.model_retrievers import (
    UUIDIdentifierModelRetriever,
    GroupIdentifierModelRetriever,
)
from api_fhir_r4.permissions import FHIRApiGroupPermissions
from api_fhir_r4.serializers import GroupSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from insuree.models import Family


class GroupViewSet(
    BaseFHIRView,
    MultiIdentifierRetrieverMixin,
    MultiIdentifierUpdateMixin,
    viewsets.ModelViewSet,
):
    retrievers = [UUIDIdentifierModelRetriever, GroupIdentifierModelRetriever]
    serializer_class = GroupSerializer
    permission_classes = (FHIRApiGroupPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        identifier = request.GET.get("identifier")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, "identifier": identifier})
        else:
            queryset = queryset.filter(validity_to__isnull=True)
        serializer = GroupSerializer(
            self.paginate_queryset(queryset), many=True, user=request.user
        )
        return self.get_paginated_response(serializer.data)

    def retrieve(self, *args, **kwargs):
        response = super().retrieve(self, *args, **kwargs)
        return response

    def get_queryset(self):
        queryset = Family.objects.all().order_by("validity_from")
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )



from unittest import mock

from api_fhir_r4.converters import PatientConverter

from fhir.resources.R4B.patient import Patient
from api_fhir_r4.tests import PatientTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToFhirTestMixin,
    ConvertToImisTestMixin,
    ConvertJsonToFhirTestMixin,
)


class PatientConverterTestCase(
    PatientTestMixin,
    ConvertToFhirTestMixin,
    ConvertToImisTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = PatientConverter
    fhir_resource = Patient
    json_repr = "test/test_patient.json"

    @mock.patch("insuree.models.Gender.objects")
    def test_to_imis_obj(self, mock_gender):
        mock_gender.get.return_value = self._TEST_GENDER
        super(PatientConverterTestCase, self).test_to_imis_obj()




from fhir.resources.R4B.organization import Organization
from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import (
    GenericFhirAPITestMixin,
    FhirApiReadTestMixin,
    LocationTestMixin,
)
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from policyholder.models import PolicyHolder
from location.test_helpers import create_test_health_facility
from api_fhir_r4.tests.utils import load_and_replace_json


class OrganisationAPITests(
    GenericFhirAPITestMixin,
    FhirApiReadTestMixin,
    LocationTestMixin,
    APITestCase,
    LogInMixin,
):
    base_url = GeneralConfiguration.get_base_url() + "Organization/"
    _test_json_path = "/test/test_organisation.json"
    _TEST_PH_CODE = "TestPHCode"
    _TEST_PH_NAME = "Test PolicyHolder"

    _test_json_path_credentials = "/test/test_login.json"

    _test_request_data_credentials = None
    _test_request_data = None
    sub_str = {}

    def get_all_bundle_entries(self, url):
        all_entries = []
        current_url = url
        while current_url:
            response = self.client.get(current_url, format="json")
            self.assertEqual(
                response.status_code, status.HTTP_200_OK, str(response.content)
            )
            bundle = self.get_bundle_from_json_response(response)
            if bundle.entry:
                all_entries.extend(bundle.entry)
            next_url = None
            if bundle.link:
                for link in bundle.link:
                    if link.relation == "next":
                        next_url = self._sanitize_next_url(link.url)
                        break
            current_url = next_url
        return all_entries

    def _sanitize_next_url(self, next_url):
        """Sanitize the next URL by replacing only the query parameters, keeping the base URL from prev_url."""
        try:
            from urllib.parse import urlparse, urlunparse, parse_qs
        except ImportError:
            from urlparse import urlparse, urlunparse, parse_qs

        # Parse the next URL to get the query parameters
        next_parsed = urlparse(next_url)
        parse_qs(next_parsed.query)

        # Combine base URL with sanitized query parameters
        sanitized_url = urlunparse(
            ("", "", next_parsed.path, "", next_parsed.query, "")
        )

        return sanitized_url

    def setUp(self):
        super(OrganisationAPITests, self).setUp()
        LocationTestMixin.setUp(self)

        self.get_or_create_user_api()
        self.create_dependencies()

        self.sub_str[self._TEST_PH_CODE] = self._TEST_PH_CODE
        self.sub_str[self._TEST_PH_NAME] = self._TEST_PH_NAME

        self._test_request_data = load_and_replace_json(
            self._test_json_path, self.sub_str
        )

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, Organization))
        self.assertEqual(updated_obj.name, self._TEST_PH_NAME)

    def update_resource(self, data):
        data["name"] = self._TEST_PH_NAME

    def create_dependencies(self):
        pass

    def test_post_should_create_correctly(self):
        self.login()
        response = self.client.post(
            self.base_url, data=self._test_request_data, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        # Verify that a PolicyHolder was created
        ph = PolicyHolder.objects.filter(code=self._TEST_PH_CODE).first()
        self.assertIsNotNone(ph)
        self.assertEqual(ph.trade_name, self._TEST_PH_NAME)

    def test_get_list_should_return_policyholders(self):
        # Create multiple policyholders
        self.login()
        # Create first policyholder
        response1 = self.client.post(
            self.base_url, data=self._test_request_data, format="json"
        )
        self.assertEqual(response1.status_code, status.HTTP_201_CREATED)

        # Create second policyholder with different data
        sub_str_2 = {
            "TestPHCode": "TestPHCode2",
            "Test PolicyHolder": "Test PolicyHolder 2",
        }
        test_data_2 = load_and_replace_json(self._test_json_path, sub_str_2)
        response2 = self.client.post(self.base_url, data=test_data_2, format="json")
        self.assertEqual(response2.status_code, status.HTTP_201_CREATED)

        # Get list of policyholders with type=bus
        bus_entries = self.get_all_bundle_entries(self.base_url + "?type=bus")
        self.assertGreaterEqual(len(bus_entries), 2)

        # Verify the policyholders are in the bundle
        found_ph1 = False
        found_ph2 = False
        for entry in bus_entries:
            if hasattr(entry.resource, "name"):
                if entry.resource.name == self._TEST_PH_NAME:
                    found_ph1 = True
                elif entry.resource.name == "Test PolicyHolder 2":
                    found_ph2 = True

        self.assertTrue(found_ph1, "First policyholder not found in list")
        self.assertTrue(found_ph2, "Second policyholder not found in list")

    def test_get_list_should_return_health_facilities(self):
        # Create multiple health facilities
        self.login()

        # Create first health facility
        create_test_health_facility(
            "HF001",
            self.test_village.parent.parent_id,
            custom_props={"name": "Test Health Facility 1"},
        )
        # Create second health facility
        create_test_health_facility(
            "HF002",
            self.test_village.parent.parent_id,
            custom_props={"name": "Test Health Facility 2"},
        )

        # Get list of health facilities with type=prov
        prov_entries = self.get_all_bundle_entries(self.base_url + "?type=prov")
        self.assertGreaterEqual(len(prov_entries), 2)

        # Verify the health facilities are in the bundle
        found_hf1 = False
        found_hf2 = False
        for entry in prov_entries:
            if hasattr(entry.resource, "name"):
                if entry.resource.name == "Test Health Facility 1":
                    found_hf1 = True
                elif entry.resource.name == "Test Health Facility 2":
                    found_hf2 = True

        self.assertTrue(found_hf1, "First health facility not found in list")
        self.assertTrue(found_hf2, "Second health facility not found in list")

    def test_get_list_should_return_insurance_organizations(self):
        # Test insurance organizations list with type=ins
        self.login()

        # Get list of insurance organizations with type=ins
        ins_entries = self.get_all_bundle_entries(self.base_url + "?type=ins")
        self.assertGreaterEqual(len(ins_entries), 1)

        # Verify the insurance organization is in the bundle
        found_ins = False
        for entry in ins_entries:
            if hasattr(entry.resource, "name"):
                # Insurance organizations should have a name
                found_ins = True
                break

        self.assertTrue(found_ins, "Insurance organization not found in list")

    def test_type_filter_should_work_correctly(self):
        # Create organizations of different types and verify type filtering
        self.login()

        # Create a policyholder
        response_ph = self.client.post(
            self.base_url, data=self._test_request_data, format="json"
        )
        self.assertEqual(response_ph.status_code, status.HTTP_201_CREATED)

        # Create health facilities
        create_test_health_facility(
            "HF001",
            self.test_village.parent.parent_id,
            custom_props={"name": "Test Health Facility 1"},
        )

        # Get all organizations without type filter
        all_entries = self.get_all_bundle_entries(self.base_url)

        # Get policyholders only (type=bus)
        bus_entries = self.get_all_bundle_entries(self.base_url + "?type=bus")

        # Get health facilities only (type=prov)
        prov_entries = self.get_all_bundle_entries(self.base_url + "?type=prov")

        # Get insurance organizations only (type=ins)
        ins_entries = self.get_all_bundle_entries(self.base_url + "?type=ins")

        # Verify that different type filters return different results
        # All should have entries
        self.assertGreater(len(all_entries), 0)
        self.assertGreater(len(bus_entries), 0)
        self.assertGreater(len(prov_entries), 0)
        self.assertGreater(len(ins_entries), 0)

        # Policyholders should be in bus but not in prov or ins
        bus_names = [
            entry.resource.name
            for entry in bus_entries
            if hasattr(entry.resource, "name")
        ]
        prov_names = [
            entry.resource.name
            for entry in prov_entries
            if hasattr(entry.resource, "name")
        ]
        ins_names = [
            entry.resource.name
            for entry in ins_entries
            if hasattr(entry.resource, "name")
        ]

        # Policyholder should be in bus results
        self.assertIn(self._TEST_PH_NAME, bus_names)

        # Policyholder should NOT be in prov or ins results
        self.assertNotIn(self._TEST_PH_NAME, prov_names)
        self.assertNotIn(self._TEST_PH_NAME, ins_names)

        # Health facility should be in prov results
        self.assertIn("Test Health Facility 1", prov_names)

        # Health facility should NOT be in bus or ins results
        self.assertNotIn("Test Health Facility 1", bus_names)
        self.assertNotIn("Test Health Facility 1", ins_names)

    def test_get_list_without_filter_should_return_all_organizations(self):
        # Test that without type filter, all organization types are returned
        self.login()

        # Create a policyholder
        response_ph = self.client.post(
            self.base_url, data=self._test_request_data, format="json"
        )
        self.assertEqual(response_ph.status_code, status.HTTP_201_CREATED)

        # Create a health facility
        create_test_health_facility(
            "HF001",
            self.test_village.parent.parent_id,
            custom_props={"name": "Test Health Facility"},
        )

        # Get all organizations without type filter
        all_entries = self.get_all_bundle_entries(self.base_url)
        self.assertGreaterEqual(
            len(all_entries), 3
        )  # At least policyholder, health facility, and insurance organization

        # Collect all organization names
        all_names = [
            entry.resource.name
            for entry in all_entries
            if hasattr(entry.resource, "name")
        ]

        # Verify that organizations from all types are present
        self.assertIn(
            self._TEST_PH_NAME,
            all_names,
            "Policyholder not found in unfiltered results",
        )
        self.assertIn(
            "Test Health Facility",
            all_names,
            "Health facility not found in unfiltered results",
        )

        # Insurance organization should also be present (from default config)
        # We can't check the exact name since it comes from config, but we know there should be at least one
        insurance_orgs = [
            name
            for name in all_names
            if name not in [self._TEST_PH_NAME, "Test Health Facility"]
        ]
        self.assertGreaterEqual(
            len(insurance_orgs),
            1,
            "Insurance organization not found in unfiltered results",
        )



from api_fhir_r4.configurations.baseApiFHIRConfiguration import IdentifierConfiguration


class R4IdentifierConfig(IdentifierConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_identifier_type_config = cfg[
            "R4_fhir_identifier_type_config"
        ]

    @classmethod
    def get_fhir_identifier_type_system(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "system",
            "https://openimis.github.io/openimis_fhir_r4_ig/CodeSystem/openimis-identifiers",
        )

    @classmethod
    def get_fhir_acsn_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_db_id_type", "ACSN"
        )

    @classmethod
    def get_fhir_id_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_db_id_type", "ACSN"
        )

    @classmethod
    def get_fhir_chfid_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_chfid_type", "SB"
        )

    @classmethod
    def get_fhir_passport_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_passport_type", "PPN"
        )

    @classmethod
    def get_fhir_facility_id_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_facility_id_type", "FI"
        )

    @classmethod
    def get_fhir_claim_admin_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_claim_admin_code_type", "FILL"
        )

    @classmethod
    def get_fhir_claim_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_claim_code_type", "Code"
        )

    @classmethod
    def get_fhir_uuid_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_db_uuid_type", "UUID"
        )

    @classmethod
    def get_fhir_generic_type_code(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_generic_code", "Code"
        )

    @classmethod
    def get_fhir_location_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_location_code_type", "Code"
        )

    @classmethod
    def get_fhir_diagnosis_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_diagnosis_code_type", "DC"
        )

    @classmethod
    def get_fhir_item_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_item_code_type", "IC"
        )

    @classmethod
    def get_fhir_service_code_type(cls):
        return cls.get_config_attribute("R4_fhir_identifier_type_config").get(
            "fhir_code_for_imis_service_code_type", "SC"
        )



# flake8: noqa

from .converter import PaymentNoticeConverter
from .mapping import (
    PaymentNoticeStatusMapping,
    PaymentNoticePaymentStatusMapping
)
from .serializer import PaymentNoticeSerializer



import logging
from policy.services import ByPolicyRequest, ByPolicyResponse, ByPolicyService
from api_fhir_r4.converters import CoverageEligibilityRequestConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer
from django.http.response import HttpResponseBase
from fhir.resources.R4B import FHIRAbstractModel
from api_fhir_r4.converters import OperationOutcomeConverter


class CoverageEligibilityRequestSerializer(BaseFHIRSerializer):

    fhirConverter = CoverageEligibilityRequestConverter
    logger = logging.getLogger(__name__)

    def to_representation(self, obj):
        if isinstance(obj, HttpResponseBase):
            return OperationOutcomeConverter.to_fhir_obj(obj).dict()
        elif isinstance(obj, FHIRAbstractModel):
            return obj.dict()
        return CoverageEligibilityRequestConverter.to_fhir_obj(obj[0], obj[1]).dict()

    def create(self, validated_data):
        request = self.context.get("request")
        eligibility_request_sp = CoverageEligibilityRequestConverter.to_imis_obj(validated_data, request.user.id)
        eligibility_request = ByPolicyRequest(chf_id=validated_data.get("chf_id"))
        try:
            response = ByPolicyService(request.user).request(eligibility_request)
        except TypeError:
            self.logger.warning(
                "The insuree with chfid `{}` is not connected with policy. "
                "The default eligibility response will be used.".format(
                    validated_data.get("chf_id")
                )
            )
            response = self.create_default_eligibility_response()
        output_response = [response, eligibility_request_sp]
        return output_response

    def create_default_eligibility_response(self):
        return ByPolicyResponse(by_insuree_request=None, items=[])



import logging


from api_fhir_r4.configurations import (
    R4LocationConfig,
    R4InvoiceConfig,
    GeneralConfiguration,
)
from api_fhir_r4.converters import (
    PatientConverter,
    BillInvoiceConverter,
    InvoiceConverter,
    HealthFacilityOrganisationConverter,
)
from api_fhir_r4.mapping.invoiceMapping import InvoiceTypeMapping, BillTypeMapping
from api_fhir_r4.subscriptions.notificationManager import (
    RestSubscriptionNotificationManager,
)
from api_fhir_r4.subscriptions.subscriptionCriteriaFilter import (
    SubscriptionCriteriaFilter,
)
from api_fhir_r4.configurations import R4ClaimConfig
from api_fhir_r4.converters import ClaimResponseConverter
from core.service_signals import ServiceSignalBindType
from core.signals import bind_service_signal
from core.utils import get_current_user
from openIMIS.openimisapps import openimis_apps
from django.contrib.auth import get_user_model


logger = logging.getLogger("openIMIS")
imis_modules = openimis_apps()
User = get_user_model()


def bind_service_signals():
    if (
        "insuree" in imis_modules
        and GeneralConfiguration.get_subscribe_insuree_signal()
    ):

        def on_insuree_create_or_update(**kwargs):
            try:
                model = kwargs.get("result", None)
                if model:
                    notify_subscribers(model, PatientConverter(), "Patient", None)
            except Exception as e:
                logger.error("Error while processing Patient Subscription", exc_info=e)

        bind_service_signal(
            "insuree_service.create_or_update",
            on_insuree_create_or_update,
            bind_type=ServiceSignalBindType.AFTER,
        )

    if "location" in imis_modules and R4LocationConfig.get_subscribe_location_signal():

        def on_hf_create_or_update(**kwargs):
            try:
                model = kwargs.get("result", None)
                if model:
                    notify_subscribers(
                        model,
                        HealthFacilityOrganisationConverter(),
                        "Organisation",
                        "bus",
                    )
            except Exception as e:
                logger.error(
                    "Error while processing Organisation Subscription", exc_info=e
                )

        bind_service_signal(
            "health_facility_service.update_or_create",
            on_hf_create_or_update,
            bind_type=ServiceSignalBindType.AFTER,
        )
    if "invoice" in imis_modules and R4InvoiceConfig.get_subscribe_invoice_signal():
        from invoice.models import Bill, Invoice

        def on_bill_create(**kwargs):
            try:
                result = kwargs.get("result", {})
                if result and result.get("success", False):
                    model_uuid = result["data"]["uuid"]
                    model = Bill.objects.get(uuid=model_uuid)
                    notify_subscribers(
                        model,
                        BillInvoiceConverter(),
                        "Invoice",
                        BillTypeMapping.invoice_type[model.subject_type.model],
                    )
            except Exception as e:
                logger.error("Error while processing Bill Subscription", exc_info=e)

        def on_invoice_create(**kwargs):
            try:
                result = kwargs.get("result", {})
                if result and result.get("success", False):
                    model_uuid = result["data"]["uuid"]
                    model = Invoice.objects.get(uuid=model_uuid)
                    notify_subscribers(
                        model,
                        InvoiceConverter(),
                        "Invoice",
                        InvoiceTypeMapping.invoice_type[model.subject_type.model],
                    )
            except Exception as e:
                logger.error("Error while processing Invoice Subscription", exc_info=e)

        bind_service_signal(
            "signal_after_invoice_module_bill_create_service",
            on_bill_create,
            bind_type=ServiceSignalBindType.AFTER,
        )
        bind_service_signal(
            "signal_after_invoice_module_invoice_create_service",
            on_invoice_create,
            bind_type=ServiceSignalBindType.AFTER,
        )

    if "claim" in imis_modules and R4ClaimConfig.get_subscribe_claim_signal():

        def on_claim_create_or_update(**kwargs):
            """
            Handles notifications when a Claim is created or updated.
            """
            try:
                model = kwargs.get("result", None)
                user = get_current_user()

                # Instantiate the correct converter (ClaimResponseConverter) with the user
                logger.info(
                    f"Processing subscription for claim {model.uuid} on behalf of user {user.username}"
                )
                converter_instance = ClaimResponseConverter(user=user)
                logger.debug(
                    f"Processing subscription for created/updated claim: {model.uuid}"
                )
                notify_subscribers(
                    model,
                    converter_instance,  # The specific converter for Claims
                    "Claim",
                    None,
                )
            except Exception as e:
                logger.error("Error while processing Claim Subscription", exc_info=e)

        bind_service_signal(
            "claim.enter_and_submit_claim",
            on_claim_create_or_update,
            bind_type=ServiceSignalBindType.AFTER,
        )


def notify_subscribers(model, converter, resource_name, resource_type_name):
    try:
        subscriptions = SubscriptionCriteriaFilter(
            model, resource_name, resource_type_name
        ).get_filtered_subscriptions()
        RestSubscriptionNotificationManager(converter).notify_subscribers_with_resource(
            model, subscriptions
        )
    except Exception as e:
        logger.error(f"Notifying subscribers failed: {e}")
        import traceback

        logger.debug(traceback.format_exc())



from fhir.resources.R4B.subscription import Subscription

from api_fhir_r4.subscriptions import SubscriptionConverter
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)
from api_fhir_r4.tests.mixin.SubscriptionTestMixin import SubscriptionTestMixin
from api_fhir_r4.utils import TimeUtils


class SubscriptionConverterTestCase(
    SubscriptionTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = SubscriptionConverter
    fhir_resource = Subscription
    json_repr = "test/test_subscription.json"

    def _load_json_repr(self):
        json_repr = super(SubscriptionConverterTestCase, self)._load_json_repr()
        json_repr["end"] = TimeUtils.str_iso_to_date(json_repr["end"])
        return json_repr



from api_fhir_r4.multiserializer.modelViewset import MultiSerializerModelViewSet
from api_fhir_r4.serializers.itemsPricelistValueSetSerializer import ItemsPricelistValueSetSerializer
from api_fhir_r4.serializers.servicesPricelistValueSetSerializer import ServicesPricelistValueSetSerializer
from api_fhir_r4.permissions import FHIRApiMedicationPermissions, FHIRApiActivityDefinitionPermissions
from medical_pricelist.models import ItemsPricelistDetail, ServicesPricelistDetail

from rest_framework.views import APIView
from api_fhir_r4.views import CsrfExemptSessionAuthentication


class PricelistValueSetViewSet(MultiSerializerModelViewSet):
    authentication_classes = [
        CsrfExemptSessionAuthentication
    ] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES

    @property
    def serializers(self):
        return {
            ItemsPricelistValueSetSerializer: (
                lambda: ItemsPricelistDetail.objects.all(),
                self._items_validator,
                (FHIRApiMedicationPermissions,),
            ),
            ServicesPricelistValueSetSerializer: (
                lambda: ServicesPricelistDetail.objects.all(),
                self._services_validator,
                (FHIRApiActivityDefinitionPermissions,),
            ),
        }

    def _items_validator(self, context):
        """Validator for items pricelist - checks if URL contains 'items-pricelist'"""
        request = context.get('request')
        if request:
            return 'items-pricelist' in request.path
        return False

    def _services_validator(self, context):
        """Validator for services pricelist - checks if URL contains 'services-pricelist'"""
        request = context.get('request')
        if request:
            return 'services-pricelist' in request.path
        return False



from api_fhir_r4.configurations import GeneralConfiguration, R4LocationConfig
from api_fhir_r4.models.imisModelEnums import ImisLocationType


class LocationTypeMapping:
    SYSTEM = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/openimis-location"
    PHYSICAL_TYPE_SYSTEM = (
        f"{GeneralConfiguration.get_system_base_url()}/CodeSystem/location-type"
    )

    PHYSICAL_TYPES_DEFINITIONS = {
        ImisLocationType.REGION.value: {
            "code": R4LocationConfig.get_fhir_code_for_region(),
            "display": "Region",
            "system": PHYSICAL_TYPE_SYSTEM,
        },
        ImisLocationType.DISTRICT.value: {
            "code": R4LocationConfig.get_fhir_code_for_district(),
            "display": "District",
            "system": PHYSICAL_TYPE_SYSTEM,
        },
        ImisLocationType.WARD.value: {
            "code": R4LocationConfig.get_fhir_code_for_ward(),
            "display": "Municipality/Ward",
            "system": PHYSICAL_TYPE_SYSTEM,
        },
        ImisLocationType.VILLAGE.value: {
            "code": R4LocationConfig.get_fhir_code_for_village(),
            "display": "Village",
            "system": PHYSICAL_TYPE_SYSTEM,
        },
    }



import copy
import uuid

from medical.models import Service
from api_fhir_r4.converters import ActivityDefinitionConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer


class ActivityDefinitionSerializer(BaseFHIRSerializer):
    fhirConverter = ActivityDefinitionConverter

    def create(self, validated_data):
        if "uuid" in validated_data.keys() and validated_data.get("uuid") is None:
            # In serializers using graphql services can't provide uuid. If uuid is provided then
            # resource is updated and not created. This check ensure UUID was provided.
            validated_data["uuid"] = uuid.uuid4()
        elif isinstance(validated_data["uuid"], str):
            validated_data["uuid"] = uuid.UUID(validated_data["uuid"])

        copied_data = copy.deepcopy(validated_data)
        if "_state" in copied_data:
            del copied_data["_state"]
        return Service.objects.create(**copied_data)

    def update(self, instance, validated_data):
        instance.code = validated_data.get("code", instance.code)
        instance.name = validated_data.get("name", instance.name)
        instance.validity_from = validated_data.get(
            "validity_from", instance.validity_from
        )
        instance.patient_category = validated_data.get(
            "patient_category", instance.patient_category
        )
        instance.category = validated_data.get("category", instance.category)
        instance.care_type = validated_data.get("care_type", instance.care_type)
        instance.type = validated_data.get("type", instance.type)
        instance.price = validated_data.get("price", instance.price)
        instance.save()
        return instance



from policyholder.models import PolicyHolder
from api_fhir_r4.converters import PolicyHolderOrganisationConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer


class PolicyHolderOrganisationSerializer(BaseFHIRSerializer):
    fhirConverter = PolicyHolderOrganisationConverter

    def create(self, validated_data):
        if PolicyHolder.objects.filter(code=validated_data["code"]).count() > 0:
            raise FHIRException(
                "Exists Organization with following code `{}`".format(
                    validated_data["code"]
                )
            )
        # Remove internal fields that shouldn't be passed to PolicyHolder constructor
        validated_data.pop("_original_state", None)
        validated_data.pop("_state", None)
        validated_data.pop("audit_user_id", None)
        request = self.context.get("request", None)
        if request and request.user:

            obj = PolicyHolder(**validated_data)
            obj.save(user=request.user)
            return obj

    def update(self, instance, validated_data):
        request = self.context.get("request", None)
        if request:
            validated_data["user_updated_id"] = request.user.id
        instance.legal_form = validated_data.get("legal_form", instance.legal_form)
        instance.phone = validated_data.get("phone", instance.phone)
        instance.email = validated_data.get("email", instance.email)
        instance.fax = validated_data.get("fax", instance.fax)
        instance.trade_name = validated_data.get("trade_name", instance.trade_name)
        instance.address = validated_data.get("address", instance.address)
        instance.bank_account = validated_data.get("bank_account", instance.address)
        instance.accountancy_account = validated_data.get(
            "accountancy_account", instance.accountancy_account
        )
        instance.user_updated_id = validated_data.get(
            "user_updated_id", instance.user_updated_id
        )
        instance.save()
        return instance



from rest_framework.request import Request
from django.utils.functional import cached_property

from api_fhir_r4.mapping.invoiceMapping import InvoiceTypeMapping, BillTypeMapping
from api_fhir_r4.mixins import (
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin,
)
from api_fhir_r4.model_retrievers import (
    CodeIdentifierModelRetriever,
    DatabaseIdentifierModelRetriever,
    UUIDIdentifierModelRetriever,
)
from api_fhir_r4.multiserializer import modelViewset
from api_fhir_r4.permissions import FHIRApiInvoicePermissions, FHIRApiBillPermissions
from api_fhir_r4.serializers import InvoiceSerializer, BillSerializer
from api_fhir_r4.views.fhir.base import BaseMultiserializerFHIRView
from api_fhir_r4.views.filters import DateUpdatedRequestParameterFilter
from invoice.models import Bill
from invoice.models import Invoice


class InvoiceViewSet(
    BaseMultiserializerFHIRView,
    modelViewset.MultiSerializerModelViewSet,
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin,
):
    retrievers = [
        UUIDIdentifierModelRetriever,
        DatabaseIdentifierModelRetriever,
        CodeIdentifierModelRetriever,
    ]
    lookup_field = "identifier"

    @cached_property
    def serializers(self):
        return {
            InvoiceSerializer: (
                self._invoice_queryset(),
                self._invoice_serializer_validator,
                (FHIRApiInvoicePermissions,),
            ),
            BillSerializer: (
                self._bill_queryset(),
                self._bill_serializer_validator,
                (FHIRApiBillPermissions,),
            ),
        }

    @classmethod
    def _invoice_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context["request"],
            get_check=lambda x: cls._get_type_from_query(x) in ("invoice", None),
            post_check=lambda x: cls._get_type_from_body(x)
            in [item["code"] for item in InvoiceTypeMapping.invoice_type],
            put_check=lambda x: cls._get_type_from_body(x)
            in [item["code"] for item in InvoiceTypeMapping.invoice_type],
        )

    @classmethod
    def _bill_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context["request"],
            get_check=lambda x: cls._get_type_from_query(x) in ("bill", None),
            post_check=lambda x: cls._get_type_from_body(x)
            in [item["code"] for item in BillTypeMapping.invoice_type],
            put_check=lambda x: cls._get_type_from_body(x)
            in [item["code"] for item in BillTypeMapping.invoice_type],
        )

    @classmethod
    def _base_request_validator_dispatcher(
        cls, request: Request, get_check, post_check, put_check
    ):
        if request.method == "GET":
            return get_check(request)
        elif request.method == "POST":
            return post_check(request)
        elif request.method == "PUT":
            return put_check(request)
        return True

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("code")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, "identifier": identifier})
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        return Invoice.objects

    def _invoice_queryset(self):
        queryset = Invoice.objects.filter(is_deleted=False).order_by("date_created")
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    def _bill_queryset(self):
        queryset = Bill.objects.filter(is_deleted=False).order_by("date_created")
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    @classmethod
    def _get_type_from_body(cls, request):
        try:
            return request.data["type"][0]["coding"][0]["code"].lower()
        except KeyError:
            return None

    @classmethod
    def _get_type_from_query(cls, request):
        try:
            return request.GET["resourceType"].lower()
        except KeyError:
            return None



# Generated by Django 3.0.14 on 2022-03-13 16:34

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ("api_fhir_r4", "0002_auto_20220313_1607"),
    ]

    operations = [
        migrations.AlterField(
            model_name="subscriptionnotificationresult",
            name="id",
            field=models.UUIDField(
                db_column="UUID",
                default=uuid.uuid4,
                editable=False,
                primary_key=True,
                serialize=False,
            ),
        ),
    ]



from core.models.user import ClaimAdmin
from django.utils.translation import gettext as _

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import (
    BaseFHIRConverter,
    PersonConverterMixin,
    ReferenceConverterMixin,
)
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.practitioner import Practitioner, PractitionerQualification
from api_fhir_r4.utils import TimeUtils, DbManagerUtils
import logging

logger = logging.getLogger("openIMIS")


class ClaimAdminPractitionerConverter(
    BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin
):

    @classmethod
    def to_fhir_obj(
        cls,
        imis_claim_admin,
        reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE,
    ):
        fhir_practitioner = Practitioner.construct()
        cls.build_fhir_pk(fhir_practitioner, imis_claim_admin, reference_type)
        cls.build_fhir_identifiers(fhir_practitioner, imis_claim_admin)
        cls.build_human_names(fhir_practitioner, imis_claim_admin)
        cls.build_fhir_birth_date(fhir_practitioner, imis_claim_admin)
        cls.build_fhir_telecom(fhir_practitioner, imis_claim_admin)
        cls.build_fhir_qualification(fhir_practitioner)
        cls.build_fhir_extension_organization(fhir_practitioner, imis_claim_admin)
        return fhir_practitioner

    @classmethod
    def to_imis_obj(cls, fhir_practitioner, audit_user_id):
        errors = []
        fhir_practitioner = Practitioner(**fhir_practitioner)
        imis_claim_admin = ClaimAdminPractitionerConverter.create_default_claim_admin(
            audit_user_id
        )
        cls.build_imis_identifiers(imis_claim_admin, fhir_practitioner, errors)
        cls.build_imis_names(imis_claim_admin, fhir_practitioner)
        cls.build_imis_birth_date(imis_claim_admin, fhir_practitioner)
        cls.build_imis_contacts(imis_claim_admin, fhir_practitioner)
        cls.check_errors(errors)
        return imis_claim_admin

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, claim_admin: ClaimAdmin):
        return claim_admin.uuid

    @classmethod
    def get_reference_obj_id(cls, claim_admin: ClaimAdmin):
        return claim_admin.id

    @classmethod
    def get_reference_obj_code(cls, claim_admin: ClaimAdmin):
        return claim_admin.code

    @classmethod
    def get_fhir_resource_type(cls):
        return Practitioner

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(
            ClaimAdmin,
            **cls.get_database_query_id_parameteres_from_reference(reference),
        )

    @classmethod
    def create_default_claim_admin(cls, audit_user_id):
        imis_claim_admin = ClaimAdmin()
        imis_claim_admin.validity_from = TimeUtils.now()
        imis_claim_admin.audit_user_id = audit_user_id
        return imis_claim_admin

    @classmethod
    def build_fhir_identifiers(cls, fhir_practitioner, imis_claim_admin):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_claim_admin)
        fhir_practitioner.identifier = identifiers

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_claim_admin):
        if imis_claim_admin.code:
            identifier = cls.build_fhir_identifier(
                imis_claim_admin.code,
                R4IdentifierConfig.get_fhir_identifier_type_system(),
                R4IdentifierConfig.get_fhir_generic_type_code(),
            )
            identifiers.append(identifier)

    @classmethod
    def build_imis_identifiers(cls, imis_claim_admin, fhir_practitioner, errors):
        value = cls.get_fhir_identifier_by_code(
            fhir_practitioner.identifier,
            R4IdentifierConfig.get_fhir_generic_type_code(),
        )
        if value:
            imis_claim_admin.code = value
        cls.valid_condition(
            imis_claim_admin.code is None, _("Missing the claim admin code"), errors
        )

    @classmethod
    def build_human_names(cls, fhir_practitioner, imis_claim_admin):
        name = cls.build_fhir_names_for_person(imis_claim_admin)
        fhir_practitioner.name = [name]

    @classmethod
    def build_imis_names(cls, imis_claim_admin, fhir_practitioner):
        names = fhir_practitioner.name
        imis_claim_admin.last_name, imis_claim_admin.other_names = (
            cls.build_imis_last_and_other_name(names)
        )

    @classmethod
    def build_fhir_extension_organization(cls, fhir_practitioner, imis_claim_admin):
        if imis_claim_admin.health_facility is not None:
            organization = cls.build_fhir_resource_reference(
                imis_claim_admin.health_facility,
                type="Organization",
                display=imis_claim_admin.health_facility.code,
                reference_type=ReferenceConverterMixin.CODE_REFERENCE_TYPE,
            )
            extension_organization = Extension.construct()
            extension_organization.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/reference"
            extension_organization.valueReference = organization
            fhir_practitioner.extension = [extension_organization]

    @classmethod
    def build_fhir_birth_date(cls, fhir_practitioner, imis_claim_admin):
        if imis_claim_admin.dob is not None:
            from core import datetime

            # check if datetime object
            if isinstance(imis_claim_admin.dob, datetime.datetime):
                fhir_practitioner.birthDate = imis_claim_admin.dob.date().isoformat()
            else:
                fhir_practitioner.birthDate = imis_claim_admin.dob.isoformat()

    @classmethod
    def build_imis_birth_date(cls, imis_claim_admin, fhir_practitioner):
        birth_date = fhir_practitioner.birthDate
        if birth_date:
            imis_claim_admin.dob = TimeUtils.str_to_date(birth_date)

    @classmethod
    def build_fhir_telecom(cls, fhir_practitioner, imis_claim_admin):
        fhir_practitioner.telecom = cls.build_fhir_telecom_for_person(
            phone=imis_claim_admin.phone, email=imis_claim_admin.email_id
        )

    @classmethod
    def build_imis_contacts(cls, imis_claim_admin, fhir_practitioner):
        imis_claim_admin.phone, imis_claim_admin.email_id = (
            cls.build_imis_phone_num_and_email(fhir_practitioner.telecom)
        )

    @classmethod
    def build_fhir_qualification(cls, fhir_practitioner):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        qualification = PractitionerQualification.construct()
        qualification.code = cls.build_codeable_concept(
            system=system, code="CA", display=_("Claim Administrator")
        )
        fhir_practitioner.qualification = [qualification]



import json
import os

from core.models import User
from fhir.resources.R4B import construct_fhir_element
from rest_framework import status

from api_fhir_r4.configurations import R4IdentifierConfig, GeneralConfiguration
from api_fhir_r4.converters import BaseFHIRConverter
from fhir.resources.R4B.bundle import Bundle
from api_fhir_r4.utils import DbManagerUtils

from core.test_helpers import create_test_interactive_user
from insuree.test_helpers import create_test_gender, create_test_profession, create_test_education, create_test_relation, create_test_confirmation_type
from location.test_helpers import create_basic_test_locations, create_test_basic_health_facility_legal_form, create_test_basic_health_facility_sub_level


class GenericFhirAPITestMixin(object):
    user = None

    @property
    def base_url(self):
        return None

    @property
    def _test_json_path(self):
        return None

    @property
    def _test_json_path_credentials(self):
        return None

    _TEST_SUPERUSER_NAME = "Admin"
    _TEST_SUPERUSER_PASS = "admin123"
    _test_request_data = None

    @classmethod
    def setUpTestData(cls):
        super(GenericFhirAPITestMixin, cls).setUpTestData()
        create_basic_test_locations()
        create_test_basic_health_facility_legal_form()
        create_test_basic_health_facility_sub_level()
        create_test_gender()
        create_test_profession()
        create_test_education()
        create_test_relation()
        create_test_confirmation_type()

    def setUp(self):
        self.user = create_test_interactive_user(username=self._TEST_SUPERUSER_NAME)
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        if self._test_json_path and self._test_request_data is None:
            json_representation = open(dir_path + self._test_json_path).read()
            self._test_request_data = json.loads(json_representation)
        if (
            self._test_json_path_credentials
            and self._test_request_data_credentials is None
        ):
            json_representation = open(
                dir_path + self._test_json_path_credentials
            ).read()
            self._test_request_data_credentials = json.loads(json_representation)

    def apply_replace_map(self, payload):
        return payload

    def initialize_auth(self):
        response = self.client.post(
            GeneralConfiguration.get_base_url() + "login/",
            data=self._test_request_data_credentials,
            format="json",
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}",
        }
        return headers

    def get_response_details(self, response_json):
        if (
            "issue" in response_json
            and len(response_json["issue"]) > 0
            and "details" in response_json["issue"][0]
            and "text" in response_json["issue"][0]["details"]
        ):
            return response_json["issue"][0]["details"]["text"]

        elif "detail" in response_json:
            return response_json["detail"]

    def login(self):
        user = DbManagerUtils.get_object_or_none(
            User, username=self._TEST_SUPERUSER_NAME
        )
        self.client.force_authenticate(user=user)

    def get_bundle_from_json_response(self, response):
        response_json = response.json()
        bundle = construct_fhir_element(response_json["resourceType"], response_json)
        self.assertTrue(isinstance(bundle, Bundle))
        return bundle

    def get_id_for_created_resource(self, response):
        result = None
        response_json = response.json()
        fhir_obj = construct_fhir_element(response_json["resourceType"], response_json)
        if hasattr(fhir_obj, "identifier"):
            result = BaseFHIRConverter.get_fhir_identifier_by_code(
                fhir_obj.identifier, R4IdentifierConfig.get_fhir_uuid_type_code()
            )
        return result

    def get_fhir_obj_from_json_response(self, response):
        fhir_obj = None
        response_json = response.json()
        if "resourceType" in response_json:
            fhir_obj = construct_fhir_element(
                response_json["resourceType"], response_json
            )

        return fhir_obj

    def test_get_should_required_login(self):
        response = self.client.get(self.base_url, data=None, format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)



from api_fhir_r4.configurations import ClaimConfiguration


class R4ClaimConfig(ClaimConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_claim_config = cfg["R4_fhir_claim_config"]

    @classmethod
    def get_fhir_claim_information_guarantee_id_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_information_guarantee_id_code", "guarantee"
        )

    @classmethod
    def get_fhir_claim_information_explanation_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_information_explanation_code", "explanation"
        )

    @classmethod
    def get_fhir_claim_item_explanation_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_explanation_code", "item_explanation"
        )

    @classmethod
    def get_fhir_claim_item_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_code", "item"
        )

    @classmethod
    def get_fhir_claim_service_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_service_code", "service"
        )

    @classmethod
    def get_fhir_claim_status_rejected_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_status_rejected_code", "rejected"
        )

    @classmethod
    def get_fhir_claim_status_entered_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_status_entered_code", "entered"
        )

    @classmethod
    def get_fhir_claim_status_checked_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_status_checked_code", "checked"
        )

    @classmethod
    def get_fhir_claim_status_processed_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_status_processed_code", "processed"
        )

    @classmethod
    def get_fhir_claim_status_valuated_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_status_valuated_code", "valuated"
        )

    @classmethod
    def get_fhir_claim_item_status_passed_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_status_passed_code", "passed"
        )

    @classmethod
    def get_fhir_claim_item_status_rejected_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_status_rejected_code", "rejected"
        )

    @classmethod
    def get_fhir_claim_item_general_adjudication_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_general_adjudication_code", "general"
        )

    @classmethod
    def get_fhir_claim_item_rejected_reason_adjudication_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_rejected_reason_adjudication_code", "rejected_reason"
        )

    @classmethod
    def get_fhir_claim_organization_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_organization_code", "openIMIS"
        )  # has to be updated when 'Organization' is created

    @classmethod
    def get_fhir_claim_attachment_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_attachment_code", "attachment"
        )

    @classmethod
    def get_fhir_claim_visit_type_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_visit_type_system", "CodeSystem/claim-visit-type"
        )

    @classmethod
    def get_fhir_claim_supporting_info_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_supporting_info_category_system",
            "CodeSystem/claim-supporting-info-category",
        )

    @classmethod
    def get_fhir_item_reference_extension_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_reference_extension_system",
            "StructureDefinition/claim-item-reference",
        )

    @classmethod
    def get_fhir_claim_item_category_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_item_category_system", "CodeSystem/claim-item-category"
        )

    @classmethod
    def get_fhir_claim_diagnosis_system(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_diagnosis_system", "CodeSystem/diagnosis-ICD10-level1"
        )

    @classmethod
    def get_fhir_claim_attachment_system(cls):
        pass

    @classmethod
    def get_allowed_fhir_claim_attachment_mime_types_regex(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_claim_allowed_mime_types_regex",
            "|".join(
                [
                    r"text/.*",  # noqa: W605
                    r"image/png",  # noqa: W605
                    r"image/jpe?g",  # noqa: W605
                    r"application/msword",  # noqa: W605
                    ".*doc.*",  # noqa: W605
                    r"application/pdf",  # noqa: W605
                ]
            ),
        )

    @classmethod
    def get_subscribe_claim_signal(cls):
        """
        Configuration flag to enable/disable notifications for the Claim resource.
        """
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "subscribe_claim_signal", True
        )



import copy

from location.gql_mutations import update_or_create_health_facility

from api_fhir_r4.converters.healthFacilityOrganisationConverter import HealthFacilityOrganisationConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer

import uuid


class HealthFacilityOrganisationSerializer(BaseFHIRSerializer):
    fhirConverter = HealthFacilityOrganisationConverter

    def create(self, validated_data):
        data = copy.deepcopy(validated_data)
        if "uuid" in validated_data and isinstance(validated_data["uuid"], str):
            validated_data["uuid"] = uuid.UUID(validated_data["uuid"])
        # UUID is automatically generated for HF model, has to be removed as create_or_update will fail.
        return self.__create_or_update(data)

    def update(self, instance, validated_data):
        validated_data["id"] = instance.id
        validated_data["uuid"] = instance.uuid

        data = copy.deepcopy(validated_data)
        return self.__create_or_update(data)

    def __create_or_update(self, validated_data):
        request = self.context.get("request")
        user = request.user
        data = copy.deepcopy(validated_data)
        if "_state" in validated_data:
            del data["_state"]

        if not data.get("offline", None):
            data["offline"] = False

        return update_or_create_health_facility(data, user)



from django.utils.translation import gettext as _
from api_fhir_r4.utils import TimeUtils
from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import (
    ClaimAdminPractitionerConverter,
    HealthFacilityOrganisationConverter,
)
from fhir.resources.R4B.contactpoint import ContactPoint
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.humanname import HumanName
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.practitioner import Practitioner, PractitionerQualification
from fhir.resources.R4B.reference import Reference

from api_fhir_r4.models.imisModelEnums import ContactPointSystem, ContactPointUse
from api_fhir_r4.tests import GenericTestMixin
from location.test_helpers import create_test_village, create_test_health_facility
from claim.test_helpers import create_test_claim_admin


class ClaimAdminPractitionerTestMixin(GenericTestMixin):

    _TEST_CLAIM_ADMIN_LAST_NAME = "Smith"
    _TEST_CLAIM_ADMIN_OTHER_NAME = "John"
    _TEST_CLAIM_ADMIN_DOB = "1990-03-24"
    _TEST_CLAIM_ADMIN_ID = 1
    _TEST_CLAIM_ADMIN_UUID = "254f6268-964b-4d8d-aa26-20081f22235e"
    _TEST_CLAIM_ADMIN_CODE = "1234abcd"
    _TEST_CLAIM_ADMIN_PHONE = "813-996-476"
    _TEST_CLAIM_ADMIN_EMAIL = "TEST@TEST.com"
    _TEST_CLAIM_ADMIN_FAX = "1-408-999 8888"
    _TEST_CLAIM_ADMIN_ADDRESS = "TEST_ADDRESS"

    _TEST_HF_ID = 90000
    _TEST_HF_UUID = "3fd89df0-a532-4df6-9b77-3cdb33b883f2"
    _TEST_HF_CODE = "92345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"

    test_claim_admin = None
    test_hf = None
    sub_str = {}

    def setUp(self):
        super(ClaimAdminPractitionerTestMixin, self).setUp()
        self.test_village = create_test_village()
        self.test_hf = self.create_test_hf()
        self.test_claim_admin = create_test_claim_admin(
            custom_props={
                "health_facility_id": self.test_hf.id,
                "code": self._TEST_CLAIM_ADMIN_CODE,
                "dob": TimeUtils.str_to_date(self._TEST_CLAIM_ADMIN_DOB),
                "phone": self._TEST_CLAIM_ADMIN_PHONE,
                "email_id": self._TEST_CLAIM_ADMIN_EMAIL,
                "last_name": self._TEST_CLAIM_ADMIN_LAST_NAME,
                "other_names": self._TEST_CLAIM_ADMIN_OTHER_NAME,
            }
        )
        self.sub_str[self._TEST_HF_UUID] = self.test_hf.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID] = self.test_claim_admin.uuid

    def create_test_hf(self):
        self.test_hf = create_test_health_facility(
            self._TEST_HF_CODE,
            self.test_village.parent.parent.id,
            custom_props={
                "name": self._TEST_HF_NAME,
                "level": self._TEST_HF_LEVEL,
                "legal_form_id": self._TEST_HF_LEGAL_FORM,
            },
        )
        self._TEST_HF_ID = self.test_hf.id
        self._TEST_HF_UUID = self.test_hf.uuid
        return self.test_hf

    def create_test_imis_instance(self, location=None, hf=None):
        return self.test_claim_admin

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_CLAIM_ADMIN_LAST_NAME, imis_obj.last_name)
        self.assertEqual(self._TEST_CLAIM_ADMIN_OTHER_NAME, imis_obj.other_names)
        self.assertEqual(self._TEST_CLAIM_ADMIN_CODE, imis_obj.code)
        self.assertEqual(
            self._TEST_CLAIM_ADMIN_DOB + "T00:00:00", imis_obj.dob.isoformat()
        )
        self.assertEqual(self._TEST_CLAIM_ADMIN_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_CLAIM_ADMIN_EMAIL, imis_obj.email_id)
        # we are not checking the extension as it's optional and the health facility
        # membership should be changed otherwise

    def create_test_fhir_instance(self):
        fhir_practitioner = Practitioner.construct()
        name = HumanName.construct()
        name.family = self._TEST_CLAIM_ADMIN_LAST_NAME
        name.given = [self._TEST_CLAIM_ADMIN_OTHER_NAME]
        name.use = "usual"
        fhir_practitioner.name = [name]
        identifiers = []
        code = ClaimAdminPractitionerConverter.build_fhir_identifier(
            self._TEST_CLAIM_ADMIN_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code(),
        )
        identifiers.append(code)
        fhir_practitioner.identifier = identifiers
        fhir_practitioner.birthDate = self._TEST_CLAIM_ADMIN_DOB
        telecom = []
        phone = ClaimAdminPractitionerConverter.build_fhir_contact_point(
            self._TEST_CLAIM_ADMIN_PHONE, ContactPointSystem.PHONE, ContactPointUse.HOME
        )
        telecom.append(phone)
        email = ClaimAdminPractitionerConverter.build_fhir_contact_point(
            self._TEST_CLAIM_ADMIN_EMAIL, ContactPointSystem.EMAIL, ContactPointUse.HOME
        )
        telecom.append(email)
        fhir_practitioner.telecom = telecom

        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        qualification = PractitionerQualification.construct()
        qualification.code = ClaimAdminPractitionerConverter.build_codeable_concept(
            system=system, code="CA", display=_("Claim Administrator")
        )
        fhir_practitioner.qualification = [qualification]

        organization_reference = Reference.construct()
        resource_type = "Organization"
        resource_id = "92345678"

        organization_reference.type = resource_type
        organization_reference.identifier = (
            HealthFacilityOrganisationConverter.build_fhir_identifier(
                self._TEST_HF_CODE,
                R4IdentifierConfig.get_fhir_identifier_type_system(),
                R4IdentifierConfig.get_fhir_generic_type_code(),
            )
        )
        organization_reference.reference = f"{resource_type}/{resource_id}"
        organization_reference.display = resource_id

        extension_organization = Extension.construct()
        extension_organization.url = (
            f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/reference"
        )
        extension_organization.valueReference = organization_reference
        fhir_practitioner.extension = [extension_organization]

        return fhir_practitioner

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(1, len(fhir_obj.name))
        human_name = fhir_obj.name[0]
        self.assertTrue(isinstance(human_name, HumanName))
        self.assertEqual(self.test_claim_admin.other_names, human_name.given[0])
        self.assertEqual(self.test_claim_admin.last_name, human_name.family)
        self.assertEqual("usual", human_name.use)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = (
                ClaimAdminPractitionerConverter.get_first_coding_from_codeable_concept(
                    identifier.type
                ).code
            )
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self.test_claim_admin.code, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code():
                self.assertEqual(str(self.test_claim_admin.uuid), identifier.value)
        self.assertEqual(self._TEST_CLAIM_ADMIN_DOB, fhir_obj.birthDate.isoformat())
        self.assertEqual(2, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            self.assertTrue(isinstance(telecom, ContactPoint))
            if telecom.system == "phone":
                self.assertEqual(self._TEST_CLAIM_ADMIN_PHONE, telecom.value)
            elif telecom.system == "email":
                self.assertEqual(self._TEST_CLAIM_ADMIN_EMAIL, telecom.value)
        self.assertEqual(1, len(fhir_obj.qualification))
        self.assertEqual("CA", fhir_obj.qualification[0].code.coding[0].code)
        self.assertEqual(
            "Claim Administrator", fhir_obj.qualification[0].code.coding[0].display
        )
        self.assertEqual("92345678", fhir_obj.extension[0].valueReference.display)



from rest_framework import status
from rest_framework.test import APITestCase
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.configurations import GeneralConfiguration
from medical.models import Item
from medical.test_helpers import create_test_item


class CodeSystemMedicalItemAPITests(GenericFhirAPITestMixin, APITestCase):
    base_url = (
        GeneralConfiguration.get_base_url() + "CodeSystem/medical-item/"
    )

    def setUp(self):
        create_test_item("D")
        super(CodeSystemMedicalItemAPITests, self).setUp()
        self._EXPECTED_COUNT = Item.objects.all().count()

    def test_get_bad_authorization(self):
        response = self.client.get(self.base_url, data=None, format="json")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_get_should_return_code_system(self):
        self.login()
        response = self.client.get(self.base_url, data=None, format="json")
        response_data = response.json()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # item_code = ""
        # for concept in response_data["concept"]:
        #     if concept["display"] == "Test item":
        #         item_code = concept["code"]
        #
        # self.assertTrue(item_code.startswith("TI-"))
        self.assertEqual(response_data["count"], self._EXPECTED_COUNT)
        self.assertEqual(response_data["name"], "MedicalItemCS")
        self.assertEqual(response_data["title"], "Medical Item")



import asyncio
import datetime
from unittest.mock import AsyncMock, MagicMock, patch
from asgiref.sync import async_to_sync
from django.test import TestCase
from api_fhir_r4.models import Subscription
from api_fhir_r4.subscriptions.notificationClient import (
    RestSubscriptionNotificationClient,
    SubscriberNotificationOutput,
)
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from aiohttp import web


class AsyncContextManagerMock(MagicMock):
    async def __aenter__(self):
        return self.aenter

    async def __aexit__(self, *args):
        pass


class TestSubscriptionNotificationClient(LogInMixin, TestCase):
    TEST_HEADERS_1 = """{"test-header": "123", "Authentication": "Bearer ABF13816"}"""
    EXPECTED_HEADER_1 = {
        "content-type": "application/json",
        "accept": "application/json",
        "test-header": "123",
        "Authentication": "Bearer ABF13816",
    }
    TEST_HEADERS_2 = """{"test-header": "123", "Authentication": "Bearer 61831FAB"}"""
    EXPECTED_HEADER_2 = {
        "content-type": "application/json",
        "accept": "application/json",
        "test-header": "123",
        "Authentication": "Bearer 61831FAB",
    }
    NOTIFICATION_CONTENT = {"notification_content": "content"}

    def setUp(self) -> None:
        super().setUp()
        self._test_user = self.get_or_create_user_api()
        self._test_subscriptions = self._create_test_subscriptions()

    @async_to_sync
    @patch("api_fhir_r4.subscriptions.notificationClient.aiohttp.ClientSession.post")
    async def test_post_should_propagate_correctly(self, session):
        session.return_value.__aenter__.return_value.json = AsyncMock(
            side_effect=[
                {"Notification": "Thanks for notification"},
                {"Notification": "Thanks for notification"},
            ]
        )
        session.return_value.__aenter__.return_value.status = 200
        sub_client = RestSubscriptionNotificationClient()
        response = await sub_client.propagate_notifications_async(
            self.NOTIFICATION_CONTENT, self._test_subscriptions
        )
        expected = [
            SubscriberNotificationOutput(self._test_subscriptions[0], True, None),
            SubscriberNotificationOutput(self._test_subscriptions[1], True, None),
        ]
        self.assertListEqual(expected, list(response))
        session.assert_any_call(
            headers=self.EXPECTED_HEADER_1,
            url="http://test-subscription-endpoint.io/post_uri/",
            data='{"notification_content":"content"}',
        )
        session.assert_any_call(
            headers=self.EXPECTED_HEADER_2,
            url="http://test-subscription-endpoint.io/post_uri/",
            data='{"notification_content":"content"}',
        )

    @async_to_sync
    @patch("api_fhir_r4.subscriptions.notificationClient.aiohttp.ClientSession.post")
    async def test_post_server_unavailable(self, session):
        server_response = {"Notification": "Server offline"}
        session.return_value.__aenter__.return_value.json = AsyncMock(
            side_effect=[server_response, server_response]
        )
        session.return_value.__aenter__.return_value.status = 503
        sub_client = RestSubscriptionNotificationClient()
        response = await sub_client.propagate_notifications_async(
            self.NOTIFICATION_CONTENT, self._test_subscriptions
        )
        expected = [
            SubscriberNotificationOutput(
                self._test_subscriptions[0], False, server_response
            ),
            SubscriberNotificationOutput(
                self._test_subscriptions[1], False, server_response
            ),
        ]
        self.assertListEqual(expected, list(response))
        session.assert_any_call(
            url="http://test-subscription-endpoint.io/post_uri/",
            headers=self.EXPECTED_HEADER_1,
            data='{"notification_content":"content"}',
        )
        session.assert_any_call(
            url="http://test-subscription-endpoint.io/post_uri/",
            headers=self.EXPECTED_HEADER_2,
            data='{"notification_content":"content"}',
        )

    def _create_test_subscriptions(self):
        _valid_subscription = [
            self._create_valid(self.TEST_HEADERS_1),
            self._create_valid(self.TEST_HEADERS_2),
        ]
        return _valid_subscription

    def _create_valid(self, headers):
        sub = Subscription(
            status=1,
            channel=0,
            endpoint="http://test-subscription-endpoint.io/post_uri/",
            headers=headers,
            expiring=datetime.datetime.now() + datetime.timedelta(days=10),
        )
        sub.save(username=self._test_user.username)
        return sub

    def _mock_coro(self):
        f = asyncio.Future()
        f.set_result(web.Response(text="Thanks for notification."))
        return f



from api_fhir_r4.converters import ValueSetConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer


class ValueSetSerializer(BaseFHIRSerializer):
    fhirConverter = ValueSetConverter

    def __init__(self, *args, **kwargs):
        self.valueSetFields = [
            "code_field",
            "display_field",
            "id",
            "name",
            "title",
            "description",
            "url",
            "codesystem_url",
            "properties",
        ]
        self.model = {}
        if "user" in kwargs:
            user = kwargs.pop("user")
        for field in self.valueSetFields:
            self.model[field] = kwargs.pop(field, None)
        if "data" in kwargs:
            self.model["data"] = kwargs.pop("data")
        else:
            self.model["data"] = {}

        super().__init__(*args, user=user, **kwargs)

    def to_representation(self, obj):
        return ValueSetConverter.to_fhir_obj(self.model, self.reference_type).dict()



import copy

from insuree.models import Family, Insuree
from api_fhir_r4.converters import GroupConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer
from insuree.services import FamilyService, InsureeService
from uuid import UUID


class GroupSerializer(BaseFHIRSerializer):
    fhirConverter = GroupConverter

    def create(self, validated_data):
        # validated_data = resolve_id_reference(Family, validated_data)
        request = self.context.get("request")
        user = request.user

        insuree_id = validated_data.get("head_insuree_id")
        members_family = validated_data.pop("members_family")

        if Family.objects.filter(head_insuree_id=insuree_id).count() > 0:
            raise FHIRException("Exists family with the provided head")
        if "uuid" in validated_data and isinstance(validated_data["uuid"], str):
            validated_data["uuid"] = UUID(validated_data["uuid"])
        insuree = Insuree.objects.get(id=insuree_id)
        copied_data = copy.deepcopy(validated_data)
        copied_data["head_insuree"] = insuree.__dict__
        copied_data["contribution"] = None
        new_family = FamilyService(user).create_or_update(copied_data)

        # assign members of family (insuree) to the family
        for mf in members_family:
            mf = mf.__dict__
            if "_state" in mf:
                del mf["_state"]
            mf["family_id"] = new_family.id
            InsureeService(user).create_or_update(mf)

        return new_family

    def update(self, instance, validated_data):
        # validated_data = resolve_id_reference(validated_data)
        # TODO: This doesn't work
        request = self.context.get("request")
        validated_data.pop("_state", None)
        members_family = validated_data.pop("members_family")
        user = request.user
        head_id = validated_data.get("head_insuree_id", None)
        if "uuid" in validated_data and isinstance(validated_data["uuid"], str):
            validated_data["uuid"] = UUID(validated_data["uuid"])
        family_uuid = validated_data.get("uuid", None)
        if head_id:
            family = Family.objects.filter(
                head_insuree_id=head_id, validity_to__isnull=True
            ).first()
            if not family:
                raise FHIRException(
                    "No family with following head id `{}`".format(head_id)
                )
        elif family_uuid:
            family = Family.objects.filter(
                uuid=family_uuid, validity_to__isnull=True
            ).first()
            if not family:
                raise FHIRException(
                    "No family with following uuid `{}`".format(head_id)
                )

        validated_data["id"] = family.id
        validated_data["uuid"] = family.uuid
        instance = FamilyService(user).create_or_update(validated_data)
        for mf in members_family:
            mf = mf.__dict__
            mf.pop("_state", None)
            mf["family_id"] = instance.id
            InsureeService(user).create_or_update(mf)

        return instance



# TODO uncomment if someone need this serializer with connection to openHIM


import logging
import os
import requests
from requests.auth import HTTPBasicAuth
from policy.services import ByInsureeRequest, ByInsureeService, ByInsureeResponse
from api_fhir_r4.converters import PolicyCoverageEligibilityRequestConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer

logger = logging.getLogger(__name__)


class PolicyCoverageEligibilityRequestSerializer(BaseFHIRSerializer):

    fhirConverter = PolicyCoverageEligibilityRequestConverter

    def create(self, validated_data):
        data = self.context.get("request").data
        data['resourceType'] = "CoverageEligibilityRequest"
        url = os.environ.get('OPEHNHIM_URL')
        technical_user = os.environ.get('OPEHNHIM_USER')
        password = os.environ.get('OPEHNHIM_PASSWORD')
        try:
            response = requests.post(
                url + 'CoverageEligibilityRequest',
                json=data,
                auth=HTTPBasicAuth(technical_user, password)
            )
            if response.status_code == 200 and len(response.json()) > 1:
                res = response.json()
                return {"items": res['insurance']['item']}
            else:
                return self.eligibility_check(validated_data.get('chf_id'))
        except Exception:
            return self.eligibility_check(validated_data.get('chf_id'))

    def eligibility_check(self, data):
        eligibility_request = ByInsureeRequest(chf_id=data)
        request = self.context.get("request")
        try:
            response = ByInsureeService(request.user).request(eligibility_request)
        except TypeError:
            self.logger.warning('The insuree with chfid `{}` is not connected with policy. '
                                'The default eligibility response will be used.'
                                .format(data.get('chfid')))
            response = self.create_default_eligibility_response()
        return response

    def create_default_eligibility_response(self):
        return ByInsureeResponse(
            by_insuree_request=None,
            items=[]
        )

    def create_eligibility_response(self, by_insuree_request, items):
        return ByInsureeResponse(
            by_insuree_request=by_insuree_request,
            items=items
        )



from api_fhir_r4.converters import MedicationConverter

from fhir.resources.R4B.medication import Medication
from api_fhir_r4.tests import MedicationTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class MedicationConverterTestCase(
    MedicationTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = MedicationConverter
    fhir_resource = Medication
    json_repr = "test/test_medication.json"



import uuid
from abc import ABC, abstractmethod
from typing import Union

from django.db.models.query import QuerySet
from django.db.models import Model
from insuree.services import validate_insuree_number
from api_fhir_r4.converters import ReferenceConverterMixin
from django.core.exceptions import ValidationError


class GenericModelRetriever(ABC):

    @property
    @abstractmethod
    def identifier_field(self) -> str:
        pass

    @property
    @abstractmethod
    def serializer_reference_type(
        self,
    ) -> Union[
        ReferenceConverterMixin.UUID_REFERENCE_TYPE,
        ReferenceConverterMixin.CODE_REFERENCE_TYPE,
        ReferenceConverterMixin.DB_ID_REFERENCE_TYPE,
    ]:
        pass

    @classmethod
    @abstractmethod
    def identifier_validator(cls, identifier_value) -> bool:
        pass

    @classmethod
    def retriever_additional_queryset_filtering(cls, queryset):
        # By default no additional changes are made in queryset
        return queryset

    @classmethod
    def get_model_object(cls, queryset: QuerySet, identifier_value) -> Model:
        filters = {}
        if cls.serializer_reference_type == "uuid_reference":
            identifier_value = uuid.UUID(str(identifier_value))
        elif hasattr(queryset.model, "validity_to"):
            filters["validity_to__isnull"] = True
        filters[cls.identifier_field] = identifier_value
        try:
            return queryset.get(**filters)
        except Exception as e:
            raise ValidationError(
                f"failed to retrieve {queryset.model.__name__} with the filter {filters}; details {e}"
            )


class UUIDIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = "uuid"
    serializer_reference_type = ReferenceConverterMixin.UUID_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return cls._is_uuid_identifier(identifier_value)

    @classmethod
    def _is_uuid_identifier(cls, identifier):
        try:
            uuid.UUID(str(identifier))
            return True
        except ValueError:
            return False


class DatabaseIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = "id"
    serializer_reference_type = ReferenceConverterMixin.DB_ID_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return isinstance(identifier_value, int) or identifier_value.isdigit()


class CodeIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = "code"
    serializer_reference_type = ReferenceConverterMixin.CODE_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return isinstance(identifier_value, str)


class CHFIdentifierModelRetriever(CodeIdentifierModelRetriever):
    identifier_field = "chf_id"

    @classmethod
    def identifier_validator(cls, identifier_value):
        # From model specification
        # Fix: Modified condition to check if validate_insuree_number returns an empty array
        # Original condition incorrectly evaluated validate_insuree_number as False when it returned an empty array []
        # New condition explicitly checks for an empty array using len(validate_insuree_number(identifier_value)) == 0
        # This ensures that a valid insuree number (returning empty array) is correctly evaluated as True
        return (
            isinstance(identifier_value, str)
            and len(validate_insuree_number(identifier_value)) == 0
        )


class GroupIdentifierModelRetriever(CHFIdentifierModelRetriever):
    identifier_field = "head_insuree_id__chf_id"



from django.urls import include, path
from drf_spectacular.views import (
    SpectacularAPIView,
    SpectacularSwaggerView,
    SpectacularRedocView,
)
from rest_framework.routers import DefaultRouter
from openIMIS.openimisapps import openimis_apps

from api_fhir_r4.views import LoginView, fhir as fhir_viewsets

imis_modules = openimis_apps()

router = DefaultRouter()
router.register(r"login", LoginView, basename="login")
router.register(
    r"Subscription", fhir_viewsets.SubscriptionViewSet, basename="Subscription_R4"
)

# register endpoint related to Product module if used
if "product" in imis_modules:
    router.register(
        r"InsurancePlan", fhir_viewsets.ProductViewSet, basename="InsurancePlan_R4"
    )

# register endpoint related to Location module if used
if "location" in imis_modules:
    router.register(r"Location", fhir_viewsets.LocationViewSet, basename="Location_R4")
    # code system for openimis organization legal form
    router.register(
        r"CodeSystem/organization-hf-legal-form",
        fhir_viewsets.CodeSystemOrganizationHFLegalFormViewSet,
        basename="CodeSystem/organization-hf-legal-form_R4",
    )
    router.register(
        r"CodeSystem/organization-hf-level",
        fhir_viewsets.CodeSystemOrganizationHFLevelViewSet,
        basename="CodeSystem/organization-hf-level_R4",
    )

# register endpoint for insuree if used
if "insuree" in imis_modules:
    router.register(r"Patient", fhir_viewsets.InsureeViewSet, basename="Patient_R4")
    router.register(r"Group", fhir_viewsets.GroupViewSet, basename="Group_R4")
    router.register(
        r"CoverageEligibilityRequest",
        fhir_viewsets.CoverageEligibilityRequestViewSet,
        basename="CoverageEligibilityRequest_R4",
    )
    # code system for openimis patient
    router.register(
        r"CodeSystem/patient-education-level",
        fhir_viewsets.CodeSystemOpenIMISPatientEducationLevelViewSet,
        basename="CodeSystem/patient-education-level_R4",
    )
    router.register(
        r"CodeSystem/patient-profession",
        fhir_viewsets.CodeSystemOpenIMISPatientProfessionViewSet,
        basename="CodeSystem/patient-profession_R4",
    )
    router.register(
        r"CodeSystem/patient-identification-type",
        fhir_viewsets.CodeSystemOpenIMISPatientIdentificationTypeViewSet,
        basename="CodeSystem/patient-identification-type_R4",
    )
    router.register(
        r"CodeSystem/patient-contact-relationship",
        fhir_viewsets.CodeSystemOpenIMISPatientRelationshipViewSet,
        basename="CodeSystem/patient-contact-relationship_R4",
    )
    # code system for openimis group
    router.register(
        r"CodeSystem/group-type",
        fhir_viewsets.CodeSystemOpenIMISGroupTypeViewSet,
        basename="CodeSystem/group-type_R4",
    )
    router.register(
        r"CodeSystem/group-confirmation-type",
        fhir_viewsets.CodeSystemOpenIMISGroupConfirmationTypeViewSet,
        basename="CodeSystem/group-confirmation-type_R4",
    )

# register endpoints related to medical module
if "medical" in imis_modules:
    router.register(
        r"Medication", fhir_viewsets.MedicationViewSet, basename="Medication_R4"
    )
    router.register(
        r"ActivityDefinition",
        fhir_viewsets.ActivityDefinitionViewSet,
        basename="ActivityDefinition_R4",
    )
    # code system for openimis medical
    router.register(
        r"CodeSystem/medical-item",
        fhir_viewsets.CodeSystemOpenIMISMedicalItemViewSet,
        basename="CodeSystem/medical-item_R4",
    )
    router.register(
        r"CodeSystem/medical-service",
        fhir_viewsets.CodeSystemOpenIMISMedicalServiceViewSet,
        basename="CodeSystem/medical-service_R4",
    )
    # value sets for openimis medical pricelists
    router.register(
        r"ValueSet/items-pricelist",
        fhir_viewsets.PricelistValueSetViewSet,
        basename="ValueSet/items-pricelist_R4",
    )
    router.register(
        r"ValueSet/services-pricelist",
        fhir_viewsets.PricelistValueSetViewSet,
        basename="ValueSet/services-pricelist_R4",
    )

# register all endpoints related to c based on c
if "claim" in imis_modules:
    router.register(r"Claim", fhir_viewsets.ClaimViewSet, basename="Claim_R4")
    router.register(
        r"ClaimResponse",
        fhir_viewsets.ClaimResponseViewSet,
        basename="ClaimResponse_R4",
    )
    router.register(
        r"PractitionerRole",
        fhir_viewsets.PractitionerRoleViewSet,
        basename="PractitionerRole_R4",
    )
    router.register(
        r"Practitioner", fhir_viewsets.PractitionerViewSet, basename="Practitioner_R4"
    )
    router.register(
        r"CommunicationRequest",
        fhir_viewsets.CommunicationRequestViewSet,
        basename="CommunicationRequest_R4",
    )
    router.register(
        r"Communication",
        fhir_viewsets.CommunicationViewSet,
        basename="Communication_R4",
    )
    # code system for openimis medication
    router.register(
        r"CodeSystem/diagnosis",
        fhir_viewsets.CodeSystemOpenIMISDiagnosisViewSet,
        basename="CodeSystem/diagnosis_R4",
    )

# register endpoint for policy if used
if "policy" in imis_modules:
    router.register(
        r"Coverage", fhir_viewsets.CoverageRequestQuerySet, basename="Coverage_R4"
    )
    router.register(r"Contract", fhir_viewsets.ContractViewSet, basename="Contract_R4")

# register endpoint for policy holder if used
if "policyholder" in imis_modules:
    router.register(
        r"Organization", fhir_viewsets.OrganisationViewSet, basename="Organisation_R4"
    )
    router.register(
        r"CodeSystem/organization-ph-legal-form",
        fhir_viewsets.CodeSystemOrganizationPHLegalFormViewSet,
        basename="CodeSystem/organization-ph-legal-form_R4",
    )
    router.register(
        r"CodeSystem/organization-ph-activity",
        fhir_viewsets.CodeSystemOrganizationPHActivityViewSet,
        basename="CodeSystem/organization-ph-activity_R4",
    )

# register endpoint for policy holder if used
if "invoice" in imis_modules:
    router.register(r"Invoice", fhir_viewsets.InvoiceViewSet, basename="Invoice_R4")
    router.register(
        r"PaymentNotice",
        fhir_viewsets.PaymentNoticeViewSet,
        basename="PaymentNotice_R4",
    )

urlpatterns = [
    path("", include(router.urls)),
    path("docs/", SpectacularAPIView.as_view(), name="docs"),
    path(
        "docs/swagger/",
        SpectacularSwaggerView.as_view(url_name="docs"),
        name="swagger-ui",
    ),
    path("docs/redoc/", SpectacularRedocView.as_view(url_name="docs"), name="redoc"),
]



from django.test import TestCase

from api_fhir_r4.models import Subscription
from api_fhir_r4.services import SubscriptionService
from api_fhir_r4.utils import TimeUtils
from core.utils import set_current_user
from core.test_helpers import create_test_interactive_user


class SubscriptionServiceTest(TestCase):
    user = None
    service = None
    create_payload = {
        "status": 1,
        "channel": 0,
        "expiring": TimeUtils.str_iso_to_date("2021-01-01T01:02:03+01:00"),
        "endpoint": "https://subscriptiontest.com/test",
        "headers": "Authentication=Bearer 12351231",
        "criteria": {
            "resource_type": "Patient",
            "chfid__startswith": "1",
        },
    }

    update_payload = {"status": 0}

    def setUp(self):
        self.user = create_test_interactive_user(username="admin_sub_service")
        set_current_user(self.user)
        self.service = SubscriptionService(user=self.user)

    def test_create(self):
        result = self.service.create(self.create_payload)
        self.assertIsInstance(result, dict)
        self.assertTrue(result["success"])
        self.assertEqual("Ok".lower(), result["message"].lower())
        created_id = result["data"]["id"]
        self.assertTrue(created_id)
        queryset = Subscription.objects.filter(id=created_id)
        self.assertEqual(queryset.count(), 1)

    def test_update(self):
        result_create = self.service.create(self.create_payload)
        created_id = result_create["data"]["id"]
        update_payload = {**self.update_payload, **{"id": created_id}}
        result = self.service.update(update_payload)
        self.assertIsInstance(result, dict)
        self.assertTrue(result["success"])
        self.assertEqual("Ok".lower(), result["message"].lower())
        changed_fields = {
            key: result["data"][key] for key in result["data"] if key in update_payload
        }
        self.assertDictEqual(update_payload, changed_fields)
        queryset = Subscription.objects.filter(id=created_id)
        self.assertEqual(queryset.count(), 1)

    def test_delete(self):
        result_create = self.service.create(self.create_payload)
        created_id = result_create["data"]["id"]
        delete_payload = {"id": created_id}
        result = self.service.delete(delete_payload)
        self.assertTrue(result["success"])
        self.assertEqual("Ok".lower(), result["message"].lower())
        queryset = Subscription.objects.filter(id=created_id, is_deleted=True)
        self.assertEqual(queryset.count(), 1)
        queryset = Subscription.objects.filter(id=created_id, is_deleted=False)
        self.assertEqual(queryset.count(), 0)



from fhir.resources.R4B.coverageeligibilityresponse import (
    CoverageEligibilityResponse as FHIRCoverageEligibilityResponse,
    CoverageEligibilityResponseInsuranceItem,
    CoverageEligibilityResponseInsurance,
    CoverageEligibilityResponseInsuranceItemBenefit,
)
from fhir.resources.R4B.money import Money
from fhir.resources.R4B.period import Period

from api_fhir_r4.configurations import (
    GeneralConfiguration,
    R4CoverageEligibilityConfiguration as Config,
)
from api_fhir_r4.converters import (
    BaseFHIRConverter,
    PatientConverter,
    ReferenceConverterMixin,
)
from api_fhir_r4.defaultConfig import DEFAULT_CFG
from api_fhir_r4.models import (
    CoverageEligibilityRequestV2 as FHIRCoverageEligibilityRequest,
)
from api_fhir_r4.utils import TimeUtils
from insuree.models import Insuree, InsureePolicy
from policy.models import Policy
from policy.services import EligibilityRequest, EligibilityService
from product.models import Product, ProductService, ProductItem
from uuid import UUID
from core.utils import get_current_user
from django.utils.translation import gettext as _


class CoverageEligibilityRequestConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(
        cls,
        coverage_eligibility_response,
        coverage_eligibility_request,
        reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE,
    ):
        fhir_response = cls.build_fhir_obligatory_fields(coverage_eligibility_request)
        insuree, fhir_response.patient = cls.build_fhir_patient(
            getattr(coverage_eligibility_request, 'chf_id', None),
            getattr(coverage_eligibility_request, 'patient_uuid', None)
        )

        if coverage_eligibility_request.policy_uuid:
            policies = Policy.objects.filter(
                uuid=coverage_eligibility_request.policy_uuid,
                insuree_policies__insuree=insuree,
                *Policy.filter_validity(prefix='insuree_policies__')
            )
        else:
            policies = Policy.objects.filter(
                product=coverage_eligibility_response.prod_id,
                insuree_policies__insuree=insuree,
                *Policy.filter_validity(prefix='insuree_policies__')
            )

        for item in policies:
            if item.status in Config.get_fhir_active_policy_status():
                cls.build_fhir_insurance(
                    fhir_response, item, coverage_eligibility_request
                )
        return fhir_response

    @classmethod
    def to_imis_obj(cls, fhir_coverage_eligibility_request, audit_user_id):
        fhir_coverage_eligibility_request = FHIRCoverageEligibilityRequest(
            **fhir_coverage_eligibility_request
        )
        chf_id = cls.build_imis_chf(fhir_coverage_eligibility_request)
        item_code, service_code = cls.build_imis_item_service(
            fhir_coverage_eligibility_request
        )
        policy_uuid = cls.build_imis_policy_uuid(fhir_coverage_eligibility_request)
        return EligibilityRequest(chf_id, service_code, item_code, policy_uuid)

    @classmethod
    def build_fhir_obligatory_fields(cls, coverage_eligibility_request):
        fhir_eligibility_response = {"status": "active", "outcome": "complete"}

        default_insurance_organisation = DEFAULT_CFG[
            "R4_fhir_insurance_organisation_config"
        ]
        resource_id = default_insurance_organisation["id"]
        reference_insurer = {"reference": f"Organization/{resource_id}"}
        fhir_eligibility_response["insurer"] = reference_insurer

        reference_patient = {
            "reference": f"Patient/{coverage_eligibility_request.chf_id}"
        }
        fhir_eligibility_response["patient"] = reference_patient

        reference_coverage_eligibility_request = {
            "reference": "CoverageEligibilityRequest"
        }
        fhir_eligibility_response["request"] = reference_coverage_eligibility_request

        fhir_eligibility_response["purpose"] = ["benefits"]
        fhir_eligibility_response["created"] = TimeUtils.date().isoformat()
        return FHIRCoverageEligibilityResponse(**fhir_eligibility_response)

    @classmethod
    def build_fhir_patient(cls, chf_id, patient_uuid):
        if chf_id:
            insuree = Insuree.objects.filter(chf_id=chf_id, *Insuree.filter_validity())
        elif patient_uuid:
            insuree = Insuree.objects.filter(uuid=patient_uuid, *Insuree.filter_validity())
        else:
            raise ValueError(_("policy.service.eligibility.insuree_id_or_uuid_missing"))
        if insuree.count() == 1:
            insuree = insuree.first()
            reference = PatientConverter.build_fhir_resource_reference(
                insuree,
                type='Patient',
                display=insuree.chf_id
            )
            return insuree, reference

    @classmethod
    def build_fhir_insurance(cls, fhir_response, item, request):
        result = CoverageEligibilityResponseInsurance.construct()
        cls.build_fhir_coverage(result, item.uuid)
        cls.build_fhir_benefit_period(result, item.start_date, item.expiry_date)
        # get the data from EligibilityService
        eligibility_request = EligibilityRequest(
            request.chf_id, request.service_code, request.item_code, item.uuid
        )
        response_eligibility_sp = EligibilityService(get_current_user()).request(
            eligibility_request
        )

        # build coverag item - product/benefit
        result.item = []
        cls.build_fhir_benefit_item_element(result, response_eligibility_sp)
        # check services and items etc
        prod_service = ProductService.objects.filter(
            product=response_eligibility_sp.prod_id, service__code=request.service_code, *ProductService.filter_validity()
        ).first()
        prod_item = ProductItem.objects.filter(
            product=response_eligibility_sp.prod_id, item__code=request.item_code, *ProductItem.filter_validity()
        ).first()
        # build coverage item - service
        if prod_service:
            cls.build_fhir_benefit_item_service_element(
                result, response_eligibility_sp, prod_service.service
            )
        # build coverage item - item
        if prod_item:
            cls.build_fhir_benefit_item_item_element(
                result, response_eligibility_sp, prod_item.item
            )
        if type(fhir_response.insurance) is not list:
            fhir_response.insurance = [result]
        else:
            fhir_response.insurance.append(result)

    @classmethod
    def build_fhir_coverage(cls, insurance, policy_uuid):
        # Due to circular dependency import has to be done inside of method
        from api_fhir_r4.converters import CoverageConverter

        policy = Policy.objects.filter(
            uuid=UUID(str(policy_uuid)), *Policy.filter_validity()
        ).first()
        reference_coverage = CoverageConverter.build_fhir_resource_reference(
            policy, type="Coverage", display=policy.uuid
        )
        insurance.coverage = reference_coverage

    @classmethod
    def build_fhir_benefit_period(cls, insurance, start_date, expiry_date):
        benefit_period = Period.construct()
        benefit_period.start = start_date
        benefit_period.end = expiry_date
        insurance.benefitPeriod = benefit_period

    @classmethod
    def build_fhir_benefit_item_element(cls, insurance, response):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system, code="benefit", display="Benefit Package"
        )
        cls.__build_item_product_name(fhir_item=item, prod_id=response.prod_id)
        item.benefit = []
        if response.total_admissions_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="admissions_left",
                display="total_admissions",
                value=response.total_admissions_left,
            )
        if response.total_consultations_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="consultations_left",
                display="total_consultations",
                value=response.total_consultations_left,
            )
        if response.total_surgeries_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="surgeries_left",
                display="total_surgeries",
                value=response.total_surgeries_left,
            )
        if response.total_deliveries_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="deliveries_left",
                display="total_deliveries",
                value=response.total_deliveries_left,
            )
        if response.total_visits_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="visits_left",
                display="total_visits",
                value=response.total_visits_left,
            )
        if response.total_antenatal_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="antenatal_left",
                display="total_antenatal",
                value=response.total_antenatal_left,
            )
        if response.hospitalization_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="hospitalization_amount",
                display="hospitalization_amount",
                value=response.hospitalization_amount_left,
            )
        if response.delivery_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="delivery_amount",
                display="delivery_amount",
                value=response.delivery_amount_left,
            )
        if response.surgery_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="surgery_amount",
                display="surgery_amount",
                value=response.surgery_amount_left,
            )
        if response.consultation_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="consultation_amount",
                display="consultation_amount",
                value=response.consultation_amount_left,
            )
        if response.antenatal_amount_left:
            cls.build_fhir_money_item_benefit_element(
                item=item,
                code="antenatal_amount",
                display="antenatal_amount",
                value=response.antenatal_amount_left,
            )
        insurance.item.append(item)

    @classmethod
    def build_fhir_benefit_item_item_element(cls, insurance, response, stored_item):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system, code="item", display="Item"
        )
        item.benefit = []
        if response.min_date_item:
            cls.build_fhir_string_item_benefit_element(
                item=item,
                code="min_date_item",
                display="Mininum date",
                value=response.min_date_item,
            )
        if response.item_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="item_left",
                display="Items left",
                value=response.service_left,
            )
        code = stored_item.code
        item.productOrService = cls.build_simple_codeable_concept(code)
        item.excluded = not response.is_service_ok
        insurance.item.append(item)

    @classmethod
    def build_fhir_benefit_item_service_element(
        cls, insurance, response, stored_service
    ):
        item = CoverageEligibilityResponseInsuranceItem.construct()
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-category"
        item.category = cls.build_codeable_concept(
            system=system, code="service", display="Service"
        )
        item.benefit = []
        if response.min_date_service:
            cls.build_fhir_string_item_benefit_element(
                item=item,
                code="min_date_item",
                display="Mininum date",
                value=response.min_date_item,
            )
        if response.item_left:
            cls.build_fhir_int_item_benefit_element(
                item=item,
                code="service_left",
                display="Services left",
                value=response.item_left,
            )
        code = stored_service.code
        item.productOrService = cls.build_simple_codeable_concept(code)
        item.excluded = not response.is_item_ok
        insurance.item.append(item)

    @classmethod
    def build_fhir_int_item_benefit_element(cls, item, code, display, value):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system, code=code, display=display
        )
        benefit.allowedUnsignedInt = value
        item.benefit.append(benefit)

    @classmethod
    def build_fhir_money_item_benefit_element(cls, item, code, display, value):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system, code=code, display=display
        )
        money_value = Money.construct()
        money_value.value = value
        benefit.allowedMoney = money_value
        item.benefit.append(benefit)

    @classmethod
    def build_fhir_string_item_benefit_element(cls, item, code, display, value):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/coverage-item-benefit-type"
        benefit = CoverageEligibilityResponseInsuranceItemBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            system=system, code=code, display=display
        )
        benefit.allowedString = value
        item.benefit.append(benefit)

    @classmethod
    def build_imis_chf(cls, fhir_coverage_eligibility_request):
        chf_id = None
        patient_reference = fhir_coverage_eligibility_request.patient
        if patient_reference:
            chf_id = PatientConverter.get_resource_id_from_reference(patient_reference)
        return chf_id

    @classmethod
    def build_imis_item_service(cls, fhir_coverage_eligibility_request):
        service_code = None
        item_code = None
        if fhir_coverage_eligibility_request.item:
            for item in fhir_coverage_eligibility_request.item:
                type_service = cls.__get_code_from_codeable_concept_by_coding_code(
                    item.category
                )
                if type_service == "item":
                    item_code = item.productOrService.text
                if type_service == "service":
                    service_code = item.productOrService.text
        return item_code, service_code

    @classmethod
    def build_imis_policy_uuid(cls, fhir_coverage_eligibility_request):
        policy_uuid = None
        coverage_reference = getattr(fhir_coverage_eligibility_request, 'coverage', None)
        if coverage_reference:
            policy_uuid = cls.get_resource_id_from_reference(coverage_reference)
        return policy_uuid

    @classmethod
    def __get_code_from_codeable_concept_by_coding_code(cls, codeable_concept):
        service_code = None
        if codeable_concept:
            coding = cls.get_first_coding_from_codeable_concept(codeable_concept)
            if coding:
                service_code = coding.code
        return service_code

    @classmethod
    def __get_coverage_data(cls, response):
        policy = (
            InsureePolicy.objects.filter(
                insuree__chf_id=response.eligibility_request.chf_id, *InsureePolicy.filter_validity()
            )
            .first()
            .policy
        )
        product = Product.objects.get(id=response.prod_id, *Product.filter_validity())
        return policy, product

    @classmethod
    def __build_item_product_name(cls, fhir_item, prod_id):
        product_queryset = Product.objects.filter(id=prod_id, *Product.filter_validity())
        product = product_queryset.first()
        fhir_item.name = product.name
        fhir_item.description = product.code



# flake8: noqa


from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer
from api_fhir_r4.serializers.patientSerializer import PatientSerializer
from api_fhir_r4.serializers.groupSerializer import GroupSerializer
from api_fhir_r4.serializers.policyHolderOrganisationSerializer import PolicyHolderOrganisationSerializer
from api_fhir_r4.serializers.locationSerializer import LocationSerializer
from api_fhir_r4.serializers.locationSiteSerializer import LocationSiteSerializer
from api_fhir_r4.serializers.claimAdminPractitionerRoleSerializer import ClaimAdminPractitionerRoleSerializer
from api_fhir_r4.serializers.claimAdminPractitionerSerializer import ClaimAdminPractitionerSerializer
from api_fhir_r4.serializers.activityDefinitionSerializer import ActivityDefinitionSerializer
from api_fhir_r4.serializers.claimSerializer import ClaimSerializer
from api_fhir_r4.serializers.healthFacilityOrganisationSerializer import HealthFacilityOrganisationSerializer
from api_fhir_r4.serializers.billSerializer import BillSerializer
from api_fhir_r4.serializers.invoiceSerializer import InvoiceSerializer
from api_fhir_r4.serializers.claimResponseSerializer import ClaimResponseSerializer
from api_fhir_r4.serializers.codeSystemSerializer import CodeSystemSerializer
from api_fhir_r4.serializers.communicationRequestSerializer import CommunicationRequestSerializer
from api_fhir_r4.serializers.communicationSerializer import CommunicationSerializer
from api_fhir_r4.serializers.contractSerializer import ContractSerializer
from api_fhir_r4.serializers.coverageSerializer import CoverageSerializer
from api_fhir_r4.serializers.coverageEligibilityRequestSerializer import CoverageEligibilityRequestSerializer
from api_fhir_r4.serializers.enrolmentOfficerPractitionerRoleSerializer import EnrolmentOfficerPractitionerRoleSerializer
from api_fhir_r4.serializers.enrolmentOfficerPractitionerSerializer import EnrolmentOfficerPractitionerSerializer
from api_fhir_r4.serializers.insuranceOrganisationSerializer import InsuranceOrganizationSerializer
from api_fhir_r4.serializers.insurancePlanSerializer import InsurancePlanSerializer
from api_fhir_r4.serializers.medicationSerializer import MedicationSerializer

from api_fhir_r4.serializers.policyCoverageEligibilityRequestSerializer import PolicyCoverageEligibilityRequestSerializer



from api_fhir_r4.converters import EnrolmentOfficerPractitionerRoleConverter
from fhir.resources.R4B.practitionerrole import PractitionerRole
from api_fhir_r4.tests import EnrolmentOfficerPractitionerRoleTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class EnrolmentOfficerPractitionerRoleConverterTestCase(
    EnrolmentOfficerPractitionerRoleTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = EnrolmentOfficerPractitionerRoleConverter
    fhir_resource = PractitionerRole
    json_repr = "test/test_enrolmentOfficerPractitionerRole.json"

    def setUp(self):
        super(EnrolmentOfficerPractitionerRoleConverterTestCase, self).setUp()
        self.sub_str[self._TEST_OFFICER_UUID] = self.test_officer.uuid



from django.utils.translation import gettext as _
from django.utils.translation import gettext
from location.models import Location

from api_fhir_r4.configurations import R4IdentifierConfig, R4LocationConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.R4B.location import Location as FHIRLocation

from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.mapping.locationMapping import LocationTypeMapping
from api_fhir_r4.utils import DbManagerUtils


class LocationConverter(BaseFHIRConverter, ReferenceConverterMixin):
    PHYSICAL_TYPES = LocationTypeMapping.PHYSICAL_TYPES_DEFINITIONS

    @classmethod
    def to_fhir_obj(
        cls, imis_location, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE
    ):
        fhir_location = FHIRLocation.construct()
        cls.build_fhir_physical_type(fhir_location, imis_location)
        cls.build_fhir_pk(fhir_location, imis_location, reference_type)
        cls.build_fhir_location_identifier(fhir_location, imis_location)
        cls.build_fhir_location_name(fhir_location, imis_location)
        cls.build_fhir_part_of(fhir_location, imis_location, reference_type)
        cls.build_fhir_status(fhir_location, imis_location)
        cls.build_fhir_mode(fhir_location)
        return fhir_location

    @classmethod
    def to_imis_obj(cls, fhir_location, audit_user_id):
        errors = []
        fhir_location = FHIRLocation(**fhir_location)
        imis_location = Location()
        cls.build_imis_location_identiftier(imis_location, fhir_location, errors)
        cls.build_imis_location_name(imis_location, fhir_location, errors)
        cls.build_imis_location_type(imis_location, fhir_location, errors)
        cls.build_imis_parent_location_id(imis_location, fhir_location, errors)
        cls.check_errors(errors)
        return imis_location

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_location_code_type()

    @classmethod
    def get_reference_obj_uuid(cls, imis_location: Location):
        return imis_location.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_location: Location):
        return imis_location.id

    @classmethod
    def get_reference_obj_code(cls, imis_location: Location):
        return imis_location.code

    @classmethod
    def get_fhir_resource_type(cls):
        return Location

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(
            Location, **cls.get_database_query_id_parameteres_from_reference(reference)
        )

    @classmethod
    def build_fhir_location_identifier(cls, fhir_location, imis_location):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_location)
        fhir_location.identifier = identifiers
        cls._validate_imis_identifiers(identifiers)

    @classmethod
    def build_fhir_location_code_identifier(cls, identifiers, imis_location):
        if imis_location is not None:
            identifier = cls.build_fhir_identifier(
                imis_location.code,
                R4IdentifierConfig.get_fhir_identifier_type_system(),
                R4IdentifierConfig.get_fhir_location_code_type(),
            )
            identifiers.append(identifier)

    @classmethod
    def build_imis_location_identiftier(cls, imis_location, fhir_location, errors):
        value = cls.get_fhir_identifier_by_code(
            fhir_location.identifier, R4IdentifierConfig.get_fhir_location_code_type()
        )
        if value:
            imis_location.code = value
        cls.valid_condition(
            imis_location.code is None, gettext("Missing location code"), errors
        )

    @classmethod
    def build_fhir_location_name(cls, fhir_location, imis_location):
        fhir_location.name = imis_location.name
        cls._validate_imis_name(imis_location)

    @classmethod
    def build_imis_location_name(cls, imis_location, fhir_location, errors):
        name = fhir_location.name
        if not cls.valid_condition(
            name is None, _("Missing location `name` attribute"), errors
        ):
            imis_location.name = name

    @classmethod
    def build_fhir_physical_type(cls, fhir_location, imis_location):
        cls._validate_physical_type(imis_location)
        system_definition = cls.PHYSICAL_TYPES.get(imis_location.type)
        fhir_location.physicalType = cls.build_codeable_concept(**system_definition)

    @classmethod
    def build_imis_location_type(cls, imis_location, fhir_location, errors):
        code = None
        try:
            code = fhir_location.physicalType.coding[0].code
            imis_code = next(
                (k for k, v in cls.PHYSICAL_TYPES.items() if v["code"] == code), None
            )
            imis_location.type = imis_code
            cls._validate_physical_type(imis_location)

        except KeyError:
            errors.append(_("Missing location physical type"))
        except FHIRException:
            errors.append(
                _(
                    "Invalid location physical type with code %(code), allowed codes are: %(types)"
                )
                % {
                    "code": code,
                    "types": [v["code"] for v in cls.PHYSICAL_TYPES.values],
                }
            )

    @classmethod
    def build_fhir_part_of(cls, fhir_location, imis_location, reference_type):
        if not cls.__is_highers_level_location(imis_location):
            fhir_location.partOf = LocationConverter.build_fhir_resource_reference(
                imis_location.parent,
                "Location",
                imis_location.parent.code,
                reference_type=reference_type,
            )

    def get_location_from_address(cls, fhir_patient_address):
        location_reference_ext = next(
            (
                ext
                for ext in fhir_patient_address.extension
                if "address-location-reference" in ext.url
            )
        )
        location = cls.get_imis_obj_by_fhir_reference(
            location_reference_ext.valueReference
        )
        if location is None and location_reference_ext:
            raise FHIRException(
                f"Invalid location reference, {location_reference_ext.valueReference} doesn't match any location."
            )
        if location is None:
            matching_locations = (
                Location.objects.filter(
                    validity_to__isnull=True,
                    name=fhir_patient_address.district,
                    parent__name=fhir_patient_address.state,
                    type="D",  # HF is expected to be at district level
                )
                .distinct()
                .all()
            )

            if matching_locations.count() != 1:
                raise FHIRException(
                    cls.__get_invalid_location_msg(
                        fhir_patient_address, matching_locations
                    )
                )
            else:
                location = matching_locations.first()
        return location

    @classmethod
    def __get_invalid_location_msg(cls, address, matching_locations):
        count = matching_locations.count()
        if count == 0:
            return _(
                f"No matching location for district {address.district}, state {address.state}."
            )
        elif count > 1:
            return _(
                f"More than one matching location district {address.district}, state {address.state}:\n"
                f"{matching_locations}."
            )

    @classmethod
    def build_imis_parent_location_id(cls, imis_location, fhir_location, errors):
        if fhir_location.partOf:
            resource = cls.get_imis_obj_by_fhir_reference(fhir_location.partOf)
            if not resource:
                errors.append(_("Invalid Location's partOf.reference"))

    @classmethod
    def _validate_imis_identifiers(cls, identifiers):
        code_identifier = cls.get_fhir_identifier_by_code(
            identifiers, cls.get_fhir_code_identifier_type()
        )
        if not code_identifier:
            raise FHIRException(
                _(
                    "Code identifier has to be present in FHIR Location identifiers, "
                    "\nidentifiers are: %(identifiers)"
                )
                % {"identifiers": identifiers}
            )

    @classmethod
    def _validate_imis_name(cls, imis_location):
        if not imis_location.name:
            raise FHIRException(
                _("Location  %(location_uuid)s without name")
                % {"uuid": imis_location.uuid}
            )

    @classmethod
    def _validate_physical_type(cls, imis_location):
        allowed_keys = cls.PHYSICAL_TYPES.keys()
        if not imis_location or (imis_location.type not in allowed_keys):

            error_msg_keys = (
                {
                    "type": imis_location.type,
                    "location": imis_location.uuid,
                    "types": allowed_keys,
                }
                if imis_location
                else {"type": None, "location": None, "types": allowed_keys}
            )
            raise FHIRException(
                _(
                    "Invalid Location's type %(type) for location %(location), "
                    "supported types are: %(types)"
                )
                % error_msg_keys
            )

    @classmethod
    def build_fhir_mode(cls, fhir_location):
        fhir_location.mode = "instance"

    @classmethod
    def build_fhir_status(cls, fhir_location, imis_location):
        if cls.__location_active(imis_location):
            fhir_location.status = R4LocationConfig.get_fhir_code_for_active()
        else:
            fhir_location.status = R4LocationConfig.get_fhir_code_for_inactive()

    @classmethod
    def __location_active(cls, imis_location):
        return imis_location.validity_to is None

    @classmethod
    def __is_highers_level_location(cls, imis_location):
        return imis_location.parent is None

    @classmethod
    def _is_code_reference(cls, location_part_of):
        if location_part_of.identifier:
            try:
                ref_type_code = location_part_of.identifier.type.coding[0].code
                return ref_type_code == R4IdentifierConfig.get_fhir_location_code_type()
            except KeyError:
                raise FHIRException(
                    _("Invalid format for Location's partOf identifier.")
                )
        else:
            # Default expected reference identifier is uuid
            return False



from api_fhir_r4.serializers.valueSetSerializer import ValueSetSerializer
from medical_pricelist.models import ItemsPricelistDetail


class ItemsPricelistValueSetSerializer(ValueSetSerializer):

    def __init__(self, *args, **kwargs):
        # Get pricelist details with related item data
        queryset = ItemsPricelistDetail.objects.select_related('item', 'items_pricelist')

        # Build data for each item with properties
        data = []
        for detail in queryset:
            item_data = {
                'code': detail.item.code,
                'display': detail.item.name,
                'price': str(detail.price_overrule) if detail.price_overrule else '0.00',
                'pricelist_name': detail.items_pricelist.name,
                'pricelist_date': str(detail.items_pricelist.pricelist_date),
            }
            data.append(item_data)

        kwargs['data'] = data
        kwargs['code_field'] = 'code'
        kwargs['display_field'] = 'display'
        kwargs['id'] = 'items-pricelist'
        kwargs['name'] = 'ItemsPricelistVS'
        kwargs['title'] = 'Medical Items Pricelist'
        kwargs['description'] = 'Medical items with pricing information from openIMIS price lists.'
        # codesystem_url will be set in to_representation
        kwargs['properties'] = ['price', 'pricelist_name', 'pricelist_date']

        super().__init__(*args, **kwargs)

    def to_representation(self, obj):
        # Set the codesystem URL dynamically
        request = self.context.get('request')
        if request:
            base_url = request.build_absolute_uri().rstrip('/ValueSet/items-pricelist')
            self.model['codesystem_url'] = f"{base_url}/CodeSystem/medical-item"
            self.model['url'] = request.build_absolute_uri()

        return super().to_representation(obj)



from api_fhir_r4.configurations import R4ClaimConfig, GeneralConfiguration
from claim.models import Claim
from django.utils.translation import gettext as _
from urllib.parse import urljoin


class ClaimPriorityMapping:
    SYSTEM = "http://terminology.hl7.org/CodeSystem/processpriority"

    fhir_priority_coding = {
        "normal": {
            "system": SYSTEM,
            "code": "normal",
            "display": "Normal",
        }
    }


class ClaimVisitTypeMapping:
    SYSTEM = urljoin(
        GeneralConfiguration.get_system_base_url(),
        R4ClaimConfig.get_fhir_claim_visit_type_system(),
    )

    fhir_claim_visit_type_coding = {
        "E": {
            "system": SYSTEM,
            "code": "E",
            "display": "Emergency",
        },
        "R": {
            "system": SYSTEM,
            "code": "R",
            "display": "Referrals",
        },
        "O": {
            "system": SYSTEM,
            "code": "O",
            "display": "Other",
        },
    }


class ClaimItemCategoryMapping:
    SYSTEM = urljoin(
        GeneralConfiguration.get_system_base_url(),
        R4ClaimConfig.get_fhir_claim_item_category_system(),
    )

    fhir_claim_item_type_coding = {
        R4ClaimConfig.get_fhir_claim_item_code(): {
            "system": SYSTEM,
            "code": R4ClaimConfig.get_fhir_claim_item_code(),
            "display": "Item",
        },
        R4ClaimConfig.get_fhir_claim_service_code(): {
            "system": SYSTEM,
            "code": R4ClaimConfig.get_fhir_claim_service_code(),
            "display": "Service",
        },
    }


class ClaimResponseMapping(object):

    claim_status_system = (
        f"{GeneralConfiguration.get_system_base_url()}CodeSystem/claim-status"
    )
    claim_status = {
        f"{Claim.STATUS_REJECTED}": R4ClaimConfig.get_fhir_claim_status_rejected_code(),
        f"{Claim.STATUS_ENTERED}": R4ClaimConfig.get_fhir_claim_status_entered_code(),
        f"{Claim.STATUS_CHECKED}": R4ClaimConfig.get_fhir_claim_status_checked_code(),
        f"{Claim.STATUS_PROCESSED}": R4ClaimConfig.get_fhir_claim_status_processed_code(),
        f"{Claim.STATUS_VALUATED}": R4ClaimConfig.get_fhir_claim_status_valuated_code(),
    }

    claim_outcome = {
        f"{Claim.STATUS_REJECTED}": _("complete"),
        f"{Claim.STATUS_ENTERED}": _("queued"),
        f"{Claim.STATUS_CHECKED}": _("partial"),
        f"{Claim.STATUS_PROCESSED}": _("partial"),
        f"{Claim.STATUS_VALUATED}": _("complete"),
    }

    visit_type_system = (
        f"{GeneralConfiguration.get_system_base_url()}CodeSystem/claim-visit-type"
    )
    visit_type = {
        "E": _("Emergency"),
        "R": _("Referrals"),
        "O": _("Other"),
    }

    rejection_reason_system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/claim-rejection-reasons"
    rejection_reason = {
        -2: _("REJECTED BY AI EVALUATION"),
        -1: _("REJECTED BY MEDICAL OFFICER"),
        0: _("ACCEPTED"),
        1: _("INVALID ITEM OR SERVICE"),
        2: _("NOT IN PRICE LIST"),
        3: _("NO PRODUCT FOUND"),
        4: _("CATEGORY LIMITATION"),
        5: _("FREQUENCY FAILURE"),
        6: _("DUPLICATED"),
        7: _("FAMILY"),
        8: _("DIAGNOSIS NOT IN LIST"),
        9: _("TARGET DATE"),
        10: _("CARE TYPE"),
        11: _("MAX HOSPITAL ADMISSIONS"),
        12: _("MAX VISITS"),
        13: _("MAX CONSULTATIONS"),
        14: _("MAX SURGERIES"),
        15: _("MAX DELIVERIES"),
        16: _("QTY OVER LIMIT"),
        17: _("WAITING PERIOD FAIL"),
        19: _("MAX ANTENATAL"),
    }



from api_fhir_r4.converters.coverageConverter import CoverageConverter
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer
from policy.models import Policy
import copy
from api_fhir_r4.exceptions import FHIRException


class CoverageSerializer(BaseFHIRSerializer):
    fhirConverter = CoverageConverter

    def create(self, validated_data):
        family = validated_data.get("family_id")
        if Policy.objects.filter(family_id=family).count() > 0:
            raise FHIRException("Exists coverage with the family  provided")
        copied_data = copy.deepcopy(validated_data)
        if "_state" in copied_data:
            del copied_data["_state"]
        return Policy.objects.create(**copied_data)

    def update(self, instance, validated_data):
        instance.status = validated_data.get("status", instance.status)
        instance.value = validated_data.get("value", instance.value)
        instance.save()
        return instance



from api_fhir_r4.converters import GroupConverter

from fhir.resources.R4B.group import Group
from api_fhir_r4.tests import GroupTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class GroupConverterTestCase(
    GroupTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = GroupConverter
    fhir_resource = Group
    json_repr = "test/test_group.json"



from rest_framework.views import APIView

from api_fhir_r4.multiserializer import MultiSerializerSerializerClass
from api_fhir_r4.paginations import FhirBundleResultsSetPagination
from api_fhir_r4.permissions import FHIRApiPermissions
from api_fhir_r4.views import CsrfExemptSessionAuthentication
from api_fhir_r4.mixins import (
    UpdateModelMixin,
    CreateModelMixin,
    DestroyModelMixin,
)


class BaseFHIRView(
    CreateModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
    APIView,
):
    user = None
    pagination_class = FhirBundleResultsSetPagination
    permission_classes = (FHIRApiPermissions,)
    authentication_classes = [
        CsrfExemptSessionAuthentication
    ] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES


class BaseMultiserializerFHIRView(
    UpdateModelMixin,
    DestroyModelMixin,
    APIView,
):
    user = None
    pagination_class = FhirBundleResultsSetPagination
    permission_classes = (FHIRApiPermissions,)
    authentication_classes = [
        CsrfExemptSessionAuthentication
    ] + APIView.settings.DEFAULT_AUTHENTICATION_CLASSES
    serializer_class = MultiSerializerSerializerClass



from api_fhir_r4.configurations.baseApiFHIRConfiguration import BaseConfiguration
from django.conf import settings


class GeneralConfiguration(BaseConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        config = cls.get_config()
        config.default_audit_user_id = cfg["default_audit_user_id"]
        config.gender_codes = cfg["gender_codes"]
        config.base_url = cfg["base_url"]
        config.default_value_of_patient_head_attribute = cfg[
            "default_value_of_patient_head_attribute"
        ]
        config.default_value_of_patient_card_issued_attribute = cfg[
            "default_value_of_patient_card_issued_attribute"
        ]
        config.default_value_of_location_offline_attribute = cfg[
            "default_value_of_location_offline_attribute"
        ]
        config.default_value_of_location_care_type = cfg[
            "default_value_of_location_care_type"
        ]
        config.default_response_page_size = cfg["default_response_page_size"]
        config.claim_rule_engine_validation = cfg["claim_rule_engine_validation"]
        config.subscribe_insuree_signal = cfg["subscribe_insuree_signal"]

    @classmethod
    def get_default_audit_user_id(cls):
        return cls.get_config_attribute("default_audit_user_id")

    @classmethod
    def get_male_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get("male", "M")

    @classmethod
    def get_female_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get("female", "F")

    @classmethod
    def get_other_gender_code(cls):
        return cls.get_config_attribute("gender_codes").get("other", "O")

    @classmethod
    def get_default_value_of_patient_head_attribute(cls):
        return cls.get_config_attribute("default_value_of_patient_head_attribute")

    @classmethod
    def get_default_value_of_patient_card_issued_attribute(cls):
        return cls.get_config_attribute(
            "default_value_of_patient_card_issued_attribute"
        )

    @classmethod
    def get_default_value_of_location_offline_attribute(cls):
        return cls.get_config_attribute("default_value_of_location_offline_attribute")

    @classmethod
    def get_default_value_of_location_care_type(cls):
        return cls.get_config_attribute("default_value_of_location_care_type")

    @classmethod
    def get_default_response_page_size(cls):
        return cls.get_config_attribute("default_response_page_size")

    @classmethod
    def get_claim_rule_engine_validation(cls):
        return cls.get_config_attribute("claim_rule_engine_validation")

    @classmethod
    def show_system(cls):
        return 1

    @classmethod
    def get_system_base_url(cls):
        return cls.get_config_attribute("base_url")

    @classmethod
    def get_host_domain(cls):
        url = cls.get_base_url()
        if url.startswith("/"):
            return f"http://{settings.SITE_URL()}"
        else:
            return ""

    @classmethod
    def get_base_url(cls):
        MODULE_NAME = "api_fhir_r4"
        site_root = settings.SITE_ROOT()
        if site_root is not None:
            base_url = "/" + site_root
        if base_url.endswith("/"):
            return base_url + MODULE_NAME + "/"
        else:
            return base_url + "/" + MODULE_NAME + "/"

    @classmethod
    def get_subscribe_insuree_signal(cls):
        return cls.get_config_attribute("subscribe_insuree_signal")



from rest_framework.exceptions import APIException

from invoice.models import PaymentInvoice, DetailPaymentInvoice
from invoice.services import PaymentInvoiceService


class PaymentNoticeSerializerUtils(object):
    _ERROR_WHILE_SAVING = "Error while saving a payment notice: %(msg)s"
    _PAYMENT_CANCEL_STATUS = [
        DetailPaymentInvoice.DetailPaymentStatus.CANCELLED,
        DetailPaymentInvoice.DetailPaymentStatus.REJECTED,
    ]

    @classmethod
    def get_result_object(cls, result):
        if result.get("success", False):
            return PaymentInvoice.objects.get(id=result["data"]["id"])
        else:
            raise APIException(
                cls._ERROR_WHILE_SAVING % {"msg": result.get("message", "Unknown")}
            )

    @classmethod
    def change_invoice_status(
        cls, payment_invoice_service: PaymentInvoiceService, payment: PaymentInvoice
    ):
        payment_detail = DetailPaymentInvoice.objects.filter(payment=payment).first()
        status = payment_detail.status
        if status == DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED:
            payment_invoice_service.payment_received(
                payment, DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED
            )
        elif status in cls._PAYMENT_CANCEL_STATUS:
            payment_invoice_service.payment_cancelled(payment)



import copy
from functools import lru_cache
from urllib.parse import urljoin

from api_fhir_r4.configurations import GeneralConfiguration, R4OrganisationConfig
from api_fhir_r4.models.imisModelEnums import ContactPointSystem
from policyholder.apps import PolicyholderConfig


class HealthFacilityOrganizationTypeMapping:
    LEGAL_FORM_CODE = "D"
    LEGAL_FORM_DISPLAY = "District organization"
    LEGAL_FORM_SYSTEM = "CodeSystem/organization-legal-form"
    LEGAL_FORM_URL = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/organization-legal-form"

    LEGAL_FORM_MAPPING = {
        "C": "Charity",
        "D": "District Organization",
        "G": "Government",
        "P": "Private organization",
    }

    ORGANIZATION_TYPE = "prov"

    EMAIL_CONTACT_POINT_SYSTEM = ContactPointSystem.EMAIL
    PHONE_CONTACT_POINT_SYSTEM = ContactPointSystem.PHONE
    FAX_CONTACT_POINT_SYSTEM = ContactPointSystem.FAX

    ADDRESS_LOCATION_REFERENCE_URL = f"""{
        GeneralConfiguration.get_system_base_url()
    }/StructureDefinition/address-location-reference"""

    CONTRACT_PURPOSE = {
        "code": "PAYOR",
        "system": "http://terminology.hl7.org/CodeSystem/contactentity-type",
    }

    LEVEL_DISPLAY_MAPPING = {"D": "Dispensary", "C": "Health Centre", "H": "Hospital"}

    LEVEL_SYSTEM = (
        f"{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-level"
    )

    TYPE_SYSTEM = f"{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-care-type"

    TYPE_DISPLAY_MAPPING = {"O": "Out-patient", "I": "In-patient", "B": "Both"}


class PolicyHolderConfigMapping:
    @classmethod
    def _get_system(cls):
        raise NotImplementedError("_get_system() not implemented")

    @classmethod
    def _get_config_mapping(cls):
        raise NotImplementedError("_get_config_mapping() not implemented")

    @classmethod
    def fhir_ph_code_system(cls, code):
        @lru_cache(maxsize=None)
        def __code_system_dict():
            system = urljoin(
                GeneralConfiguration.get_system_base_url(), cls._get_system()
            )
            dict_ = {}
            for activity in copy.deepcopy(cls._get_config_mapping()):
                activity["system"] = system
                dict_[int(activity["code"])] = activity
            return dict_

        return __code_system_dict()[code]


class PolicyHolderOrganisationLegalFormMapping(PolicyHolderConfigMapping):
    @classmethod
    def _get_system(cls):
        return R4OrganisationConfig.get_fhir_ph_organisation_legal_form_code_system()

    @classmethod
    def _get_config_mapping(cls):
        return PolicyholderConfig.policyholder_legal_form


class PolicyHolderOrganisationActivityMapping(PolicyHolderConfigMapping):
    @classmethod
    def _get_system(cls):
        return R4OrganisationConfig.get_fhir_ph_organisation_activity_code_system()

    @classmethod
    def _get_config_mapping(cls):
        return PolicyholderConfig.policyholder_activity



from api_fhir_r4.configurations import CoverageConfiguration


class R4CoverageConfig(CoverageConfiguration):

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_identifier_type_config = cfg["R4_fhir_coverage_config"]
        cls.get_config().R4_fhir_contract_config = cfg["R4_fhir_contract_config"]

    @classmethod
    def get_status_policy_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_policy_status", "Policy"
        )

    @classmethod
    def get_status_executable_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_executable_status", "Executable"
        )

    @classmethod
    def get_status_terminated_code(cls):

        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_terminated_status", "Terminated"
        )

    @classmethod
    def get_status_disputed_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_disputed_status", "Disputed"
        )

    @classmethod
    def get_status_offered_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_offered_status", "Offered"
        )

    @classmethod
    def get_status_renewed_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_renewed_status", "Renewed"
        )

    @classmethod
    def get_signer_eo_type_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_eo_signer_type", "EnrolmentOfficer"
        )

    @classmethod
    def get_signer_head_type_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_head_signer_type", "HeadOfFamily"
        )

    @classmethod
    def get_offer_insuree_role_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_insuree_role", "Insuree"
        )

    @classmethod
    def get_offer_dependant_role_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_contract_dependant_role", "Dependant"
        )

    @classmethod
    def get_organization_code(cls):
        return cls.get_config_attribute("R4_fhir_contract_config").get(
            "fhir_organization_code", "openIMIS"
        )

    @classmethod
    def get_family_reference_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_family_refereence_code", "FamilyReference"
        )

    @classmethod
    def get_status_idle_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_status_idle_code", "Idle"
        )

    @classmethod
    def get_status_cancelled_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_status_cancelled_code", "cancelled"
        )

    @classmethod
    def get_status_draft_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_status_draft_code", "draft"
        )

    @classmethod
    def get_status_active_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_status_active_code", "active"
        )

    @classmethod
    def get_status_suspended_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_status_suspended_code", "suspended"
        )

    @classmethod
    def get_status_expired_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_status_expired_code", "Expired"
        )

    @classmethod
    def get_item_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_item_code", "item"
        )

    @classmethod
    def get_service_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_service_code", "service"
        )

    @classmethod
    def get_practitioner_role_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_practitioner_role_code", "Practitioner"
        )

    @classmethod
    def get_product_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_product_code", "Product"
        )

    @classmethod
    def get_enroll_date_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_enroll_date_code", "EnrollDate"
        )

    @classmethod
    def get_effective_date_code(cls):
        return cls.get_config_attribute("R4_fhir_claim_config").get(
            "fhir_effective_date_code", "EffectiveDate"
        )



from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, ListModelMixin
from api_fhir_r4.model_retrievers import (
    UUIDIdentifierModelRetriever,
    CodeIdentifierModelRetriever,
)
from api_fhir_r4.permissions import FHIRApiClaimPermissions
from api_fhir_r4.serializers import ClaimResponseSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from claim.models import Claim


class ClaimResponseViewSet(
    BaseFHIRView, ListModelMixin, MultiIdentifierRetrieverMixin, GenericViewSet
):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]
    serializer_class = ClaimResponseSerializer
    permission_classes = (FHIRApiClaimPermissions,)

    def get_queryset(self):
        queryset = Claim.get_queryset(None, self.request.user).order_by("validity_from")
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(
            queryset
        )



from api_fhir_r4.configurations import PaymentNoticeConfiguration


class R4PaymentNoticeConfig(PaymentNoticeConfiguration):
    _config = "R4_fhir_payment_notice_config"

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_payment_notice_config = cfg[
            "R4_fhir_payment_notice_config"
        ]

    @classmethod
    def get_fhir_payment_notice_status_active(cls):
        return cls.get_config_attribute("R4_fhir_payment_notice_config").get(
            "fhir_payment_notice_status_active", "active"
        )

    @classmethod
    def get_fhir_payment_notice_status_cancelled(cls):
        return cls.get_config_attribute("R4_fhir_payment_notice_config").get(
            "fhir_payment_notice_status_cancelled", "cancelled"
        )

    @classmethod
    def get_fhir_payment_notice_status_draft(cls):
        return cls.get_config_attribute("R4_fhir_payment_notice_config").get(
            "fhir_payment_notice_status_draft", "draft"
        )

    @classmethod
    def get_fhir_payment_notice_status_entered_in_error(cls):
        return cls.get_config_attribute("R4_fhir_payment_notice_config").get(
            "fhir_payment_notice_status_entered_in_error", "entered-in-error"
        )

    @classmethod
    def get_fhir_payment_notice_payment_status_paid(cls):
        return cls.get_config_attribute("R4_fhir_payment_notice_config").get(
            "fhir_payment_notice_payment_status_paid", "paid"
        )

    @classmethod
    def get_fhir_payment_notice_payment_status_cleared(cls):
        return cls.get_config_attribute("R4_fhir_payment_notice_config").get(
            "fhir_payment_notice_payment_status_cleared", "cleared"
        )



from api_fhir_r4.configurations import SubscriptionConfiguration


class R4SubscriptionConfig(SubscriptionConfiguration):
    _config = "R4_fhir_subscription_config"

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_config().R4_fhir_subscription_config = cfg[
            "R4_fhir_subscription_config"
        ]

    @classmethod
    def get_fhir_sub_search_perms(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "fhir_sub_search_perms", ["158001"]
        )

    @classmethod
    def get_fhir_sub_create_perms(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "fhir_sub_create_perms", ["158002"]
        )

    @classmethod
    def get_fhir_sub_update_perms(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "fhir_sub_update_perms", ["158003"]
        )

    @classmethod
    def get_fhir_sub_delete_perms(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "fhir_sub_delete_perms", ["158004"]
        )

    @classmethod
    def get_fhir_subscription_channel_rest_hook(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "fhir_sub_channel_rest_hook", "rest_hook"
        )

    @classmethod
    def get_fhir_subscription_status_off(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "fhir_sub_status_off", "off"
        )

    @classmethod
    def get_fhir_subscription_status_active(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "fhir_sub_status_active", "active"
        )

    @classmethod
    def get_fhir_sub_criteria_key_resource(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "get_fhir_sub_criteria_key_resource", "resource"
        )

    @classmethod
    def get_fhir_sub_criteria_key_resource_type(cls):
        return cls.get_config_attribute("R4_fhir_subscription_config").get(
            "get_fhir_sub_criteria_key_resource_type", "resource_type"
        )



import copy
import uuid

from core.models.user import ClaimAdmin

from api_fhir_r4.converters import ClaimAdminPractitionerConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers.baseSerializer import BaseFHIRSerializer


class ClaimAdminPractitionerSerializer(BaseFHIRSerializer):

    fhirConverter = ClaimAdminPractitionerConverter

    def create(self, validated_data):
        code = validated_data.get("code")
        if "uuid" in validated_data.keys() and validated_data.get("uuid") is None:
            # In serializers using graphql services can't provide uuid. If uuid is provided then
            # resource is updated and not created. This check ensure UUID was provided.
            validated_data["uuid"] = uuid.uuid4()
        elif "uuid" in validated_data and isinstance(validated_data["uuid"], str):
            validated_data["uuid"] = uuid.UUID(validated_data["uuid"])
        if ClaimAdmin.objects.filter(code=code).count() > 0:
            raise FHIRException(
                "Exists practitioner with following code `{}`".format(code)
            )
        copied_data = copy.deepcopy(validated_data)
        if "_state" in copied_data:
            del copied_data["_state"]
        return ClaimAdmin.objects.create(**copied_data)

    def update(self, instance, validated_data):
        if "uuid" in validated_data and isinstance(validated_data["uuid"], str):
            validated_data["uuid"] = uuid.UUID(validated_data["uuid"])
        instance.code = validated_data.get("code", instance.code)
        instance.last_name = validated_data.get("last_name", instance.last_name)
        instance.other_names = validated_data.get("other_names", instance.other_names)
        instance.dob = validated_data.get("dob", instance.dob)
        instance.phone = validated_data.get("phone", instance.phone)
        instance.email_id = validated_data.get("email_id", instance.email_id)
        instance.audit_user_id = self.get_audit_user_id()
        instance.save()
        return instance



from api_fhir_r4.converters import EnrolmentOfficerPractitionerConverter

from fhir.resources.R4B.practitioner import Practitioner
from api_fhir_r4.tests import EnrolmentOfficerPractitionerTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class EnrolmentOfficerPractitionerConverterTestCase(
    EnrolmentOfficerPractitionerTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = EnrolmentOfficerPractitionerConverter
    fhir_resource = Practitioner
    json_repr = "test/test_enrolmentOfficerPractitioner.json"



# Generated by Django 3.2.16 on 2023-05-04 11:56
import logging

from django.db import migrations
from core.utils import insert_role_right_for_system

logger = logging.getLogger(__name__)

IMIS_ADMIN_ROLE_IS_SYSTEM = 64
SUB_SEARCH_ROLE_RIGHT_ID = 158001
SUB_CREATE_ROLE_RIGHT_ID = 158002
SUB_UPDATE_ROLE_RIGHT_ID = 158003
SUB_DELETE_ROLE_RIGHT_ID = 158004


def add_rights(apps, schema_editor):
    """
    Add subscription CRUD permission to the IMIS Administrator.
    """
    insert_role_right_for_system(
        IMIS_ADMIN_ROLE_IS_SYSTEM, SUB_SEARCH_ROLE_RIGHT_ID, apps
    )
    insert_role_right_for_system(
        IMIS_ADMIN_ROLE_IS_SYSTEM, SUB_CREATE_ROLE_RIGHT_ID, apps
    )
    insert_role_right_for_system(
        IMIS_ADMIN_ROLE_IS_SYSTEM, SUB_UPDATE_ROLE_RIGHT_ID, apps
    )
    insert_role_right_for_system(
        IMIS_ADMIN_ROLE_IS_SYSTEM, SUB_DELETE_ROLE_RIGHT_ID, apps
    )


def remove_rights(apps, schema_editor):
    """
    Remove subscription CRUD permissions to the IMIS Administrator.
    """
    RoleRight = apps.get_model("core", "RoleRight")

    RoleRight.objects.filter(
        role__is_system=IMIS_ADMIN_ROLE_IS_SYSTEM,
        right_id__in=[
            SUB_CREATE_ROLE_RIGHT_ID,
            SUB_DELETE_ROLE_RIGHT_ID,
            SUB_SEARCH_ROLE_RIGHT_ID,
            SUB_UPDATE_ROLE_RIGHT_ID,
        ],
        validity_to__isnull=True,
    ).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("api_fhir_r4", "0005_auto_20221012_0818"),
        ("core", "0015_missing_roles"),
    ]

    operations = [
        migrations.RunPython(add_rights, remove_rights),
    ]



# flake8: noqa
from api_fhir_r4.multiserializer.serializerClass import MultiSerializerSerializerClass
from api_fhir_r4.multiserializer import mixins
from api_fhir_r4.multiserializer import modelViewset



from rest_framework.exceptions import APIException
from django.utils.translation import gettext


class FHIRException(APIException):
    def __init__(self, message):
        super(FHIRException, self).__init__(message)


class FHIRRequestProcessException(FHIRException):
    def __init__(self, errors):
        base_massage = gettext(
            "The request cannot be processed due to the following issues:\n"
        )
        message = base_massage + ",\n".join(errors)
        super(FHIRRequestProcessException, self).__init__(message)


class InvalidAttributeError(FHIRException):
    def __init__(self, attr, property_type):
        msg = gettext(
            "The attribute named '{}' is not a valid property for '{}'."
        ).format(attr, property_type)
        super(InvalidAttributeError, self).__init__(msg)


class PropertyError(FHIRException):
    def __init__(self, message):
        super(PropertyError, self).__init__(message)


class PropertyMaxSizeError(PropertyError):
    def __init__(self, definition):
        message = gettext("The max size was exceeded for property {} [{}..{}]").format(
            definition.name, definition.count_min, definition.count_max
        )
        super(PropertyMaxSizeError, self).__init__(message)


class PropertyTypeError(Exception):
    def __init__(self, local_type, description):
        msg = gettext("Expected '{}' but got '{}' for '{}' property").format(
            description.type, local_type, description.name
        )
        super(PropertyTypeError, self).__init__(msg)


class UnsupportedFormatError(Exception):
    def __init__(self, data_format):
        message = gettext("The format '{}' is not supported").format(data_format)
        super(UnsupportedFormatError, self).__init__(message)



from api_fhir_r4.converters.claimConverter import ClaimConverter
from api_fhir_r4.models import ClaimV2 as Claim
from api_fhir_r4.tests import ClaimTestMixin
from api_fhir_r4.tests.mixin import (
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
)


class ClaimConverterTestCase(
    ClaimTestMixin,
    ConvertToImisTestMixin,
    ConvertToFhirTestMixin,
    ConvertJsonToFhirTestMixin,
):
    converter = ClaimConverter
    fhir_resource = Claim
    json_repr = "test/test_claim.json"
